from typing import Tuple, NamedTuple, Union
import math
import numpy as np
from collections import namedtuple

Coordinate = namedtuple('Coordinate', ('x', 'y'))

AltiusViewportDefn = namedtuple('AltiusViewportDefn',
                                ('rawimage_nx',                   # Number of X pixels on raw detector, default 1504
                                 'rawimage_ny',                   # Number of Y pixels on raw detector, default 2000
                                 'viewport_x0',                   # bottom left corner X index of Read-out viewport_definition in detecteor pixels, default 163
                                 'viewport_y0',                   # bottom left corner Y index of read-out viewport_definition in detector pixels, default 410
                                 'viewport_nx',                   # width of read-out viewport_definition in detector pixels, default 1176
                                 'viewport_ny',                   # height of read-out viewport_definition in detector pixels, default 1176
                                 'science_x0',                    # bottom left corner X index of science window in read-out viewport_definition pixels, default 0
                                 'science_y0',                    # bottom left corner Y index of science window in read-out viewport_definition pixels, default 0
                                 'science_xbin',                  # pixel binning of the science window in the X direction in read-out viewport_definition pixels, default 3
                                 'science_ybin',                  # pixel binning of the science window in the Y direction in read-out viewport_definition pixels, default 3
                                 'science_nx',                    # Number of X (binned) science pixels, default 392
                                 'science_ny',                    # Number of Y (binned) science pixels, default 392
                                 'deconv_x0',                     # bottom left corner X index of deconvolution window in science window pixels, 0
                                 'deconv_y0',                     # bottom left corner Y index of deconvolution window in science window pixels, 0
                                 'deconv_xbin',                   # pixel binning of the deconvolution window in the X direction in science window pixels, default 28
                                 'deconv_ybin',                   # pixel binning of the deconvolution window in the Y direction in science window pixels, default 1
                                 'deconv_nx',                     # Number of X deconvolution window pixels, default 14
                                 'deconv_ny'),                     # Number of Y deconvolution window pixels, default 392
                                defaults=(1504, 2000, 163, 410, 1176, 1176, 0, 0, 3, 3, 392, 392, 0, 0, 28, 1, 14, 392))


# ------------------------------------------------------------------------------
#           Class AltiusDetectorPixelToAngle
# ------------------------------------------------------------------------------
class AltiusDetectorPixelToAngle:

    def __init__(self,
                 radians_per_raw_pixel: float,
                 boresight_raw_pixel: Coordinate,
                 boresight_target_height: float = 50000.0,
                 earth_radius: float = 6360000.0,
                 satellite_height: float =691000.0):

        self._radians_per_raw_pixel = radians_per_raw_pixel
        self._boresight_target_height = boresight_target_height
        self._Re = earth_radius
        self._Rs = earth_radius + satellite_height
        self._boresight_raw_pixel = boresight_raw_pixel

    # ------------------------------------------------------------------------------
    #           radians_per_detector_pixel
    # ------------------------------------------------------------------------------
    @property
    def radians_per_detector_pixel(self):
        return self._radians_per_raw_pixel

    # ------------------------------------------------------------------------------
    #           raw_pixel_to_delta_radians
    # ------------------------------------------------------------------------------
    def pixel_to_radians(self, pixy: np.ndarray) -> np.ndarray:

        dtheta = (pixy - self._boresight_raw_pixel.y) * self._radians_per_raw_pixel
        return dtheta

    # ------------------------------------------------------------------------------
    #           delta_radians_to_raw_pixel
    # ------------------------------------------------------------------------------
    def delta_radians_to_raw_pixel_y(self, delta_radians: np.ndarray) -> np.ndarray:

        pixy = delta_radians / self._radians_per_raw_pixel + self._boresight_raw_pixel.y
        dtheta = (pixy - self._boresight_raw_pixel.y) * self._radians_per_raw_pixel
        return dtheta

    # ------------------------------------------------------------------------------
    #           raw_pixel_to_delta_radians
    # ------------------------------------------------------------------------------
    def raw_pixel_x_to_delta_radians(self, pixx: np.ndarray) -> np.ndarray:

        dtheta = (pixx - self._boresight_raw_pixel.x) * self._radians_per_raw_pixel
        return dtheta

    # ------------------------------------------------------------------------------
    #           delta_radians_to_raw_pixel
    # ------------------------------------------------------------------------------
    def delta_radians_to_raw_pixel_x(self, delta_radians: np.ndarray) -> np.ndarray:

        pixx = delta_radians / self._radians_per_raw_pixel + self._boresight_raw_pixel.x
        dtheta = (pixx - self._boresight_raw_pixel.x) * self._radians_per_raw_pixel
        return dtheta

    # ------------------------------------------------------------------------------
    #           delta_radians_to_height
    # ------------------------------------------------------------------------------
    def delta_radians_to_height(self, deltaradians: np.ndarray) -> np.ndarray:
        theta0 = math.acos((self._Re + self._boresight_target_height) / self._Rs)
        theta = theta0 - deltaradians
        h = np.cos(theta) * self._Rs - self._Re
        return h

    # ------------------------------------------------------------------------------
    #           delta_radians_to_height
    # ------------------------------------------------------------------------------
    def height_to_delta_radians(self, height_meters: np.ndarray) -> np.ndarray:
        theta = np.acos((self._Re + height_meters) / self._Rs)
        theta0 = math.acos((self._Re + self._boresight_target_height) / self._Rs)
        dtheta = theta0 - theta
        return dtheta

    # ------------------------------------------------------------------------------
    #           raw_pixel_to_altitude
    # ------------------------------------------------------------------------------
    def raw_pixel_to_height(self, pixy: Union[float, np.ndarray]) -> np.ndarray:

        dtheta = self.pixel_to_radians(pixy)
        h = self.delta_radians_to_height(dtheta)
        return h

    # ------------------------------------------------------------------------------
    #           raw_pixel_to_altitude
    # ------------------------------------------------------------------------------
    def height_to_raw_pixel(self, h: Union[float, np.ndarray]) -> np.ndarray:

        dtheta = self.height_to_delta_radians(h)
        pixy = self.delta_radians_to_raw_pixel_y(dtheta)
        return pixy


# ------------------------------------------------------------------------------
#           class AltiusViewport:
# ------------------------------------------------------------------------------
class AltiusViewport:

    def __init__(self, viewport_definition: AltiusViewportDefn, radians_per_raw_pixel: float, boresight_raw_pixel: Coordinate = None, **kwargs):

        self._viewport_defn = viewport_definition
        if boresight_raw_pixel is None:
            boresight_raw_pixel = Coordinate(x=viewport_definition.viewport_x0 + 0.5 * viewport_definition.viewport_nx,
                                             y=viewport_definition.viewport_y0 + 0.5 * viewport_definition.viewport_ny)

        self._pixel_to_angle = AltiusDetectorPixelToAngle(radians_per_raw_pixel, boresight_raw_pixel, **kwargs)

    # ------------------------------------------------------------------------------
    #           pixel_to_angle
    # ------------------------------------------------------------------------------
    @property
    def pixel_to_angle(self):
        return self._pixel_to_angle

    # ------------------------------------------------------------------------------
    #           viewport_definition
    # ------------------------------------------------------------------------------
    @property
    def viewport_definition(self) -> AltiusViewportDefn:
        return self._viewport_defn

    # ------------------------------------------------------------------------------
    #           science_window_fovy
    # ------------------------------------------------------------------------------

    def science_window_fovy(self):
        fov_y = (self._viewport_defn.science_ny * self._viewport_defn.science_ybin) * self._pixel_to_angle.radians_per_detector_pixel
        return fov_y

    # ------------------------------------------------------------------------------
    #           science_pixel_to_raw
    # ------------------------------------------------------------------------------
    def science_pixel_y_to_raw(self, science_y: np.ndarray):

        iy = science_y * self._viewport_defn.science_ybin + self._viewport_defn.science_ybin // 2 + self._viewport_defn.science_y0 + self._viewport_defn.viewport_y0
        return iy

    # ------------------------------------------------------------------------------
    #           convert_raw_image_to_science
    # ------------------------------------------------------------------------------
    def convert_raw_image_to_science(self, raw_image: np.ndarray) -> np.ndarray:

        self.verify_raw_image_shape(raw_image)
        ix0 = self._viewport_defn.science_x0 + self._viewport_defn.viewport_x0
        iy0 = self._viewport_defn.science_y0 + self._viewport_defn.viewport_y0
        numsci_x = self._viewport_defn.science_nx
        numsci_y = self._viewport_defn.science_ny
        binx = self._viewport_defn.science_xbin
        biny = self._viewport_defn.science_ybin
        ix1 = ix0 + numsci_x * binx
        iy1 = iy0 + numsci_y * biny
        raw0 = raw_image[iy0:iy1, ix0:ix1]
        raw1 = raw0.reshape([numsci_y, biny, numsci_x, binx])
        sciimage = np.sum(raw1, axis=(1, 3))
        self.verify_science_image_shape(sciimage)
        return sciimage

    # ------------------------------------------------------------------------------
    #           science_to_deconv_image
    # ------------------------------------------------------------------------------
    def convert_science_image_to_deconv(self, science_image: np.ndarray) -> np.ndarray:

        self.verify_science_image_shape(science_image)
        ix0 = self._viewport_defn.deconv_x0
        iy0 = self._viewport_defn.deconv_y0
        numdeconv_x = self._viewport_defn.deconv_nx
        numdeconv_y = self._viewport_defn.deconv_ny
        binx = self._viewport_defn.deconv_xbin
        biny = self._viewport_defn.deconv_ybin
        ix1 = ix0 + numdeconv_x * binx
        iy1 = iy0 + numdeconv_y * biny
        raw0 = science_image[iy0:iy1, ix0:ix1]
        raw1 = raw0.reshape([numdeconv_y, biny, numdeconv_x, binx])
        deconvimage = np.sum(raw1, axis=(1, 3))
        self.verify_deconv_image_shape(deconvimage)
        return deconvimage

    # ------------------------------------------------------------------------------
    #           verify_science_image_shape
    # ------------------------------------------------------------------------------
    def verify_science_image_shape(self, science_image: np.ndarray):

        assert (science_image.shape[0] == self._viewport_defn.science_ny) and \
               (science_image.shape[1] == self._viewport_defn.science_nx), \
               Exception('The science image [{},{}] does not match the required viewport_definition sizes [{},{}]'.format(science_image.shape[0], science_image.shape[1], self._viewport_defn.science_ny, self._viewport_defn.science_nx))
        assert science_image.flags['C_CONTIGUOUS'], ValueError('The science image must be C_CONTIGUOUS')

    # ------------------------------------------------------------------------------
    #           verify_raw_image_shape
    # ------------------------------------------------------------------------------
    def verify_raw_image_shape(self, raw_image: np.ndarray):

        assert (raw_image.shape[0] == self._viewport_defn.rawimage_ny) and \
               (raw_image.shape[1] == self._viewport_defn.rawimage_nx), \
               ValueError('The raw image [{},{}] does not match the required viewport_definition sizes [{},{}]'.format(raw_image.shape[0], raw_image.shape[1], self._viewport_defn.rawimage_ny, self._viewport_defn.rawimage_nx))
        assert raw_image.flags['C_CONTIGUOUS'], ValueError('The raw image must be C_CONTIGUOUS')

    # ------------------------------------------------------------------------------
    #           verify_deconv_image_shape
    # ------------------------------------------------------------------------------
    def verify_deconv_image_shape(self, deconv_image: np.ndarray):

        assert (deconv_image.shape[0] == self._viewport_defn.deconv_ny) and \
               (deconv_image.shape[1] == self._viewport_defn.deconv_nx), \
               ValueError('The deconvolved image [{},{}] does not match the required viewport_definition sizes [{},{}]'.format(deconv_image.shape[0], deconv_image.shape[1], self._viewport_defn.deconv_ny, self._viewport_defn.deconv_nx))


# ------------------------------------------------------------------------------
#           AltiusViewport_UV
# ------------------------------------------------------------------------------
class AltiusViewport_UV(AltiusViewport):

    def __init__(self):
        viewport = AltiusViewportDefn(rawimage_nx=1504,
                                      rawimage_ny=2000,
                                      viewport_x0=494,
                                      viewport_y0=742,
                                      viewport_nx=510,
                                      viewport_ny=513,
                                      science_x0=0,
                                      science_y0=0,
                                      science_xbin=3,
                                      science_ybin=3,
                                      science_nx=170,
                                      science_ny=171,
                                      deconv_x0=0,
                                      deconv_y0=0,
                                      deconv_xbin=34,
                                      deconv_ybin=1,
                                      deconv_nx=5,
                                      deconv_ny=171)
        radians_per_raw_pixel = 0.034 / 513.0
        super().__init__(viewport, radians_per_raw_pixel)
