from typing import Any, Dict
from .window_mapping import WindowMappingHierarchy, WindowMappingType, AltiusWindowSpecs, Coords
from .window_transforms import WindowTransforms
from .configuration_key import AltiusConfigurationKey
from .load_specs import load_specs


# ------------------------------------------------------------------------------
#           Class DetectorSpecs
# ------------------------------------------------------------------------------
class DetectorSpecs:
    """
    A simple class with a list of attributes taken form the yaml file
    """

    def __init__(self, configkey: AltiusConfigurationKey, window_config_name='standard'):

        specs = load_specs('detectors.yaml')
        yaml_specs = specs[configkey.channel]
        self._boresight_location: Coords = Coords(y=yaml_specs['boresight_location'][0], x=yaml_specs['boresight_location'][1], windowtype=WindowMappingType.WINDOW_DETECTOR)
        self._read_out_noise: float = yaml_specs['read-out-noise']
        self._electrons_per_dn: float = yaml_specs['electrons_per_dn']
        self._all_window_specs = {}
        self._window_specs_list: AltiusWindowSpecs = None
        self._windowhierarchy: WindowMappingHierarchy = None
        self._window_transforms: WindowTransforms = None
        self._windowing_config = ''

        for name, keys in yaml_specs['window'].items():
            window_specs_list = self._fetch_window_specs_list(keys)
            windowhierarchy = WindowMappingHierarchy(window_specs_list)
            self._all_window_specs[name] = windowhierarchy, window_specs_list
        self.select_window_config(window_config_name)

    @property
    def window_hierarchy(self) -> WindowMappingHierarchy:
        """
        The internal object that has the hierachy of windows going from the near-out-field window to the L1C window, with all windows
        inbetween.

        Returns
        -------
        WindowMappingHierarchy
        """
        return self._windowhierarchy

    @property
    def boresight_location(self) -> Coords:
        """
        The (Y,X) boresight location in detector window pixels as a Coords object The center of pixels is on the 0.5 pixel value

        Returns
        -------
        Coords
        """
        return self._boresight_location

    @property
    def read_out_noise(self) -> float:
        """
        The readout noise entry in the yaml file. Value is in electrons

        Returns
        --------
        float
        """
        return self._read_out_noise

    @property
    def electrons_per_dn(self) -> float:
        """
        The number of electrons per digital number

        Returns
        -------
        float
        """
        return self._electrons_per_dn

    @property
    def window_specs(self) -> AltiusWindowSpecs:
        """
         A named tuple that stores all of the windows specs for all of the different windows used by this
         detector/channel

        Returns
        --------
        AltiusWindowSpecs
        """
        return self._window_specs_list

    @property
    def window_transforms(self) -> WindowTransforms:
        "The internal window transform object"
        return self._window_transforms

    # ------------------------------------------------------------------------------
    #           window_specs
    # ------------------------------------------------------------------------------
    def _fetch_window_specs_list(self, w: Dict[str, Any]) -> AltiusWindowSpecs:
        """
        Loads the named tuple of the window specifications used for the 5 ALTIUS windows..
        """

        windowspecs = AltiusWindowSpecs(FULLCONVx=w['fullconv_window'][1],
                                        FULLCONVy=w['fullconv_window'][0],
                                        NOFSx=w['noof_window'][3],
                                        NOFSy=w['noof_window'][2],
                                        NOOFx=w['noof_window'][1],
                                        NOOFy=w['noof_window'][0],
                                        DOFSx=w['detector_window'][3],
                                        DOFSy=w['detector_window'][2],
                                        Dx=w['detector_window'][1],
                                        Dy=w['detector_window'][0],
                                        SOFSx=w['science_window'][3],
                                        SOFSy=w['science_window'][2],
                                        Sx=w['science_window'][1],
                                        Sy=w['science_window'][0],
                                        LBINx=w['l1b_window'][3],
                                        LBINy=w['l1b_window'][2],
                                        Lx=w['l1b_window'][1],
                                        Ly=w['l1b_window'][0],
                                        KBINx=w['l1c_window'][3],
                                        KBINy=w['l1c_window'][2],
                                        Kx=w['l1c_window'][1],
                                        Ky=w['l1c_window'][0])
        return windowspecs

    # ------------------------------------------------------------------------------
    #           select_window_config
    # ------------------------------------------------------------------------------
    def select_window_config(self, name: str):
        """
        Selects the windowing parameters used by the detector object. The default value loaded at initialization is "standard"

        Parameters
        ----------
        name: str
            The name of the new windowing configuration. This will map to a configuration key under the window
            keyword in the detector.yaml file.
        """

        self._windowhierarchy, self._window_specs_list = self._all_window_specs[name]
        self._windowing_config = name
        self._window_transforms = WindowTransforms(self._window_specs_list)
