from typing import Dict, List, Any, Union
from collections import namedtuple
from datetime import datetime
import numbers
import numpy as np
from skconfig import Configuration
import sktime
from .window_mapping import coord_from_tuple, Coords
from .load_specs import load_specs
from .configuration_key import AltiusConfigurationKey


# ------------------------------------------------------------------------------
#           AltiusWindowSpecs
# ------------------------------------------------------------------------------
#: Attributes
#: ----------
#: S: float
#:     The power law term in the modified harvey distribution. This is a positive value typically close to 2.0 (ie between 1.5 and 3.5). It is not negative
#: b0: float
#:     The amplitude term. Or the value ofthe maofied harvey stray light at :math:`\theta = 0.0`.
#: l: float
#:     The shoulder angle. The power law behaviour is modified at angles smaller than the shoulder to value so the function
#:     behaves properly at zero.
HarveyScatteringCoeffs = namedtuple('HarveyScatteringCoeffs', ['S', 'l', 'b0'])


CentralFieldScatteringCoeffs = namedtuple('CentralFieldScatteringCoeffs', ['a0', 'k', 'b0', 'ellipticity_y'])


# ------------------------------------------------------------------------------
#           PSFSpecsConvolutionMatrixEntry
# ------------------------------------------------------------------------------
class PSFSpecsConvolutionMatrixEntry:

    def __init__(self, yaml: Configuration):

        self._detector_windows = {}
        self._identification: str = None
        self._infield_matrix: Dict[str, Any] = None
        self._horizontal_noof_matrix: Dict[str, Any] = None

        for id in yaml['detector_window'].keys():
            infield = yaml.subkey('detector_window/{}/infield_matrix'.format(id))
            horizontal_noof_matrix = yaml.subkey("detector_window/{}/horizontal_noof_matrix".format(id))
            self._detector_windows[id] = (infield, horizontal_noof_matrix)
        self.set_detector_window('standard')

    # ------------------------------------------------------------------------------
    #           set_detector_window
    # ------------------------------------------------------------------------------
    def set_detector_window(self, name: str):
        self._infield_matrix, self._horizontal_noof_matrix = self._detector_windows[name]
        self._identification = name

    @property
    def detector_window_identification(self):
        return self._identification

    @property
    def infield_matrix_settings(self):
        return self._infield_matrix

    @property
    def horizontal_noof_matrix_settings(self):
        return self._horizontal_noof_matrix


# ------------------------------------------------------------------------------
#           PSFSpecsDataFileEntry
# ------------------------------------------------------------------------------
class PSFSpecsDataFileEntry:

    def __init__(self, wavelen: float, utc: datetime, yaml: Configuration):

        self._wavelen = wavelen
        self._utc = utc
        self._center_location = coord_from_tuple(yaml['center_location'])
        self._psf_filename = yaml.as_pathname('psf_filename')

    @property
    def center_of_psf(self) -> Coords:
        """
        The location of the center of the PSF in detector window coordinates.
        """
        return self._center_location

    @property
    def utc(self) -> datetime:
        """
        The timestamp of this PSF image entry.
        """
        return self._utc

    @property
    def wavelen(self):
        "The wavelength of this PSF image entry"
        return self._wavelen

    @property
    def psf_filename(self):
        "The name of the source PSF filename"
        return self._psf_filename


# ------------------------------------------------------------------------------
#           Class PSFSpecsWavelengthEntry
# ------------------------------------------------------------------------------
class PSFSpecsWavelengthEntry:

    def __init__(self, wavelen, utc: datetime, yaml: Configuration):

        far_field_psf = yaml.subkey('far_field_psf')
        harvey_settings = far_field_psf['harvey_straylight_params'] if far_field_psf is not None else None
        near_field_psf = yaml.subkey('near_field_psf')
        near_field_coeffs = near_field_psf['central_field_scattering_coeffs'] if near_field_psf is not None else None
        infield_settings = yaml.subkey('convolution_matrix')

        self._wavelen = wavelen
        self._utc = utc
        self._convolution_matrix = PSFSpecsConvolutionMatrixEntry(infield_settings)
        self._far_field_radial_fit_range: List[float] = far_field_psf['radial_angle_range']
        self._far_field_bad_spots: List[float] = far_field_psf['bad_spots']
        self._far_field_harvey_coeffs = HarveyScatteringCoeffs(S=harvey_settings['S'], l=harvey_settings['l'], b0=harvey_settings['b0']) if harvey_settings is not None else None
        self._near_field_psf = CentralFieldScatteringCoeffs(a0=near_field_coeffs['a0'], k=near_field_coeffs['k'], b0=['b0'], ellipticity_y=near_field_psf['ellipticity_y']) if near_field_coeffs is not None else None
        self._psf_image_location_array = []
        self._psf_interpolation_coeffs_filename = 'undefined'
        self._psf_vignetting_filename = 'undefined'
        self._high_altitude_proxy_stray_light_algorithm: Dict[str, Any] = yaml['high_altitude_proxy_stray_light_algorithm']

        self._psf_interpolation_coeffs_filename = yaml['interpolation_coeffs']
        self._psf_vignetting_filename = yaml['vignetting_array']
        psf_data_files = yaml.subkey('psf_data_files')
        for i in range(len(psf_data_files.values)):
            datentry = PSFSpecsDataFileEntry(wavelen, utc, psf_data_files.subkey(i))
            self._psf_image_location_array.append(datentry)

    def image_ids(self):
        return self._psf_image_location_array.keys()

    @property
    def wavelen(self):
        return self._wavelen

    @property
    def utc(self):
        return self._utc

    @property
    def high_altitude_proxy_stray_light_algorithm(self) -> Dict[str, Any]:
        return self._high_altitude_proxy_stray_light_algorithm

    @property
    def convolution_matrix_settings(self) -> PSFSpecsConvolutionMatrixEntry:
        """
        The settings used for matrix convolution generation
        """
        return self._convolution_matrix

    @property
    def far_field_angular_fit_range(self) -> List[float]:
        return self._far_field_radial_fit_range

    @property
    def far_field_harvey_coeffs(self) -> HarveyScatteringCoeffs:
        """
        Modified Harvey parameters that describe the far-field stray light component
        """
        return self._far_field_harvey_coeffs

    @property
    def central_field_scattering_coeffs(self) -> CentralFieldScatteringCoeffs:
        """
        Parameters that describe the stray light in the central field region
        """
        return self._near_field_psf

    @property
    def psf_interpolation_coeffs_filename(self):
        """
        The name of the file that stores the indicies and coefficients used to interpolate the array of PSF images to any point on the
         detector window surface. This field will have the value 'undefined' if no interpolation table exists.
        """
        return self._psf_interpolation_coeffs_filename

    @property
    def psf_vignetting_array_filename(self):
        return self._psf_vignetting_filename

    # ------------------------------------------------------------------------------
    #           select_image
    # ------------------------------------------------------------------------------
    def select_image_location(self, idcode: int):
        return self._psf_image_location_array[idcode]

    # ------------------------------------------------------------------------------
    #           __getitem__
    # ------------------------------------------------------------------------------
    def __getitem__(self, image_location_id: int) -> PSFSpecsDataFileEntry:
        """
        Overloads the indexing operator to return settings for a given wavelength

        Parameters
        -----------
        image_location_id: int
            The id number of the requested image location. This parameter is used to tag the location of the PSF central
            lobe in the detector window. The value of 0 is typically used for PSF images at or very near to the center
            of the detector window.

        Returns
        -------
        PSFSpecsDataFileEntry
            The PSF information for this location. It may be None if the requested location tag is invalid

        """
        setting = self.select_image_location(image_location_id)
        assert setting is not None, "No dictionary entry found for the requested image location id {}".format(image_location_id)
        return setting


# ------------------------------------------------------------------------------
#           PSFSpecsDateEntry
# ------------------------------------------------------------------------------
class PSFSpecsDateEntry:
    """
    Stores the PSF entries for a given date. This is a list of wavelength based entries which are stored as a dictionary
    of wavelength values. By default the first wavelength entry is set as the defat
    """
    def __init__(self, utc: datetime, yaml: Configuration):

        self._utc = utc
        self._wavelength_entries: Dict[float, PSFSpecsWavelengthEntry] = {}
        self._current_wavelength: float = np.nan
        self._current_entry: PSFSpecsWavelengthEntry = None
        self._description: str = yaml['description']

        wavelenentries = yaml.subkey('wavelengths')
        for wavelen in wavelenentries.values:
            wavelen_float = float(wavelen)
            yamlentry = wavelenentries.subkey(wavelen)
            wavelen_entry = PSFSpecsWavelengthEntry(float(wavelen), self._utc, yamlentry)
            self._wavelength_entries[wavelen_float] = wavelen_entry
        self.select_first_wavelength()

    @property
    def utc(self) -> datetime:
        return self._utc

    # ------------------------------------------------------------------------------
    #           first_wavelength_entry
    # ------------------------------------------------------------------------------
    def select_first_wavelength(self):
        wavelen = next(iter(self._wavelength_entries))
        self.select_wavelength(wavelen)

    # ------------------------------------------------------------------------------
    #           select_wavelength
    # ------------------------------------------------------------------------------
    def select_wavelength(self, wavelen: float) -> PSFSpecsWavelengthEntry:
        entry = self._wavelength_entries.get(wavelen)
        self._current_entry = entry
        self._current_wavelength = wavelen if entry is not None else np.nan
        return self._current_entry

    # ------------------------------------------------------------------------------
    #           __getitem__
    # ------------------------------------------------------------------------------
    def __getitem__(self, wavelength: float) -> PSFSpecsWavelengthEntry:
        """
        Overloads the indexing operator to return settings for a given wavelength

        Parameters
        -----------
        wavelength: float
            The wavelength in nm of the required setting. This must identically match a setting given in the psf.yaml file
            for the current channel and date.

        Returns
        -------
        PSFSpecsWavelengthEntry
            The PSF information for the wavelength entry. It may be None if the requested wavelength is invalid

        """
        setting = self.select_wavelength(wavelength)
        assert setting is not None, "No dictionary entry found for the requested wavelength {}".format(wavelength)
        return setting


# ------------------------------------------------------------------------------
#           PSFSpecs
# ------------------------------------------------------------------------------
class PSFSpecs:
    """
    Stores the PSF specifications for a given channel. This is a list of date based settings. By default the last date
    setting is used as the default.
    """

    def __init__(self, configkey: AltiusConfigurationKey):
        """

        Parameters
        ----------
        channel: str
            Specifies the Altius channel. It must be one of "vis", "uv", or "nir"

        date_value: str, datetime, np.datetime64 or float
            [Optional]. Specifies the UT date and time of the required PSF entry. The date and time must exactly match the
            date and time stored in the configuration file. It can be any representation of time that can be coerced by package ``sktime``
            into a valid datetime object. By default it is None and the last entry is loaded in.
        """

        self._date_entries: List[PSFSpecsDateEntry] = []                                        # The list of date entries
        self._current_date: PSFSpecsDateEntry = None                                            # The current active date entry
        config = load_specs('psf.yaml')

        yaml_specs = config.subkey(configkey.channel)
        for datesettings in yaml_specs.values:
            currentyaml = yaml_specs.subkey(datesettings)
            dateentry = PSFSpecsDateEntry(datesettings, currentyaml)
            self._date_entries.append(dateentry)
        if configkey.datestr is None:
            self.select_last_date()
        else:
            self.select_date(configkey.datestr)

    # ------------------------------------------------------------------------------
    #           select_last_date
    # ------------------------------------------------------------------------------
    def select_last_date(self):
        self._current_date = self._date_entries[-1]

    # ------------------------------------------------------------------------------
    #           select_date
    # ------------------------------------------------------------------------------
    def select_date(self, datestr):

        utc = sktime.ut_to_datetime(datestr)
        ok = self._current_date.utc == utc if self._current_date is not None else False
        if not ok:
            index = None
            for i, value in enumerate(self._date_entries):
                if value.utc == utc:
                    index = i
                    break
            if index is not None:
                self._current_date = self._date_entries[index]
            else:
                raise "UTC entry {} was not found in the PSF configuration".format(datestr)
        return self._current_date

    # ------------------------------------------------------------------------------
    #           __getitem__
    # ------------------------------------------------------------------------------
    def __getitem__(self, wavelength: Union[float, str]) -> Union[PSFSpecsWavelengthEntry, PSFSpecsDateEntry]:
        """
        Overloads the indexing operator. This can be used to change the current date of the settings if a date is given
        (str, datetime, np.datetime64, float over 2000) or it can be used to change the current wavelength in
        the current date settings (number given between 200 and 2000)

        Parameters
        -----------
        wavelength: float, datetimem np.datetime64, str
            If this this a number between 200 and 2000 then it assumes this is a wavelength and returns the corresponding
            wavelength setting of the current date entry.  Otherwise it assumes parameter ``wavelength`` represents a UTC
            value and sets the current date entry usingthe UTC value.

        Returns
        -------
        PSFSpecsWavelengthEntry or PSFSpecsDateEntry
            Returns the wavelength entry if indexing a wavelength. Else selects the request date entry returns it as the the current date entry

        """
        if (isinstance(wavelength, numbers.Number)) and wavelength < 2000:
            w = float(wavelength)
            assert self._current_date is not None, "The current active PSF configuration date is not set"
            assert (w > 200) and (w < 2000), "The wavelength must be specified between 2 and 2000"
            answer = self._current_date[w]
        else:
            utc = sktime.ut_to_datetime(wavelength)
            self.select_date(utc)
            answer = self._current_date
        return answer
