from typing import Tuple
from collections import namedtuple
import numpy as np
import scipy.io
import scipy.optimize
import matplotlib.pyplot as plt
from ..configuration import OpticsSpecs
from ..configuration.window_mapping import WindowMappingType, Coords
from ..configuration.psf_configuration import CentralFieldScatteringCoeffs, HarveyScatteringCoeffs


# ------------------------------------------------------------------------------
#           Class CentralFieldPSF
# ------------------------------------------------------------------------------
class CentralFieldPSF:
    """
    A class used to model the near-field point spread function. This is the fairly large central lobe near to the on-axis signal.
    This lobe has an elliptical shape due to the tilting of mirrors in the aft-optics. This code is currently hand coded  for
    the VIS channel.

    This fits function :math:`\\exp\\left[A_0\\exp(-k\\theta') + B_0\\right]` to the near field region where :math:`\\theta'` is
    a angle modified for elliptical variation.
    """
    def __init__(self):

        self._center = Coords(x=751.6139323491205, y=999.9313305514934, windowtype=WindowMappingType.WINDOW_DETECTOR)
        self._rads_per_pix_x = 2.2606382978723404e-05
        self._rads_per_pix_y = 2.2617686170212765e-05
        self._elliptical_y_factor = 0.88586387 

    # ------------------------------------------------------------------------------
    #           get_dx_and_dy_distances_of_pixels
    # ------------------------------------------------------------------------------
    def dx_and_dy_distances_of_pixels(self, psf: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns an array giving the distance of each pixel from the center of the PSF
        :param psf:
        :return:
        """

        Ny = psf.shape[0]
        Nx = psf.shape[1]
        xpos = ((self._center.x - 0.5) - np.arange(Nx)).reshape([1, Nx])
        ypos = ((self._center.y - 0.5) - np.arange(Ny)).reshape([Ny, 1])
        dx = np.tile(xpos, [Ny, 1])
        dy = np.tile(ypos, [1, Nx])
        return dx, dy

    # ------------------------------------------------------------------------------
    #           get_elliptically_corrected_angles
    # ------------------------------------------------------------------------------
    
    def get_modified_angle_of_pixels(self, dx: np.ndarray, dy: np.ndarray) -> np.ndarray:
        """
        Corrects true angles across the detcetor surface to *modified* angles that account for the
        ellipticity of the near field PSF.

        Parameters
        ----------
        dx: np.ndarray
            The delta X offset array
        dy: np.ndarray
            The delta Y offset array
        Returns
        -------
        np.ndarray
            The array of modified angles.
        """
        ddy = dy * (self._rads_per_pix_y * self._elliptical_y_factor)
        ddx = dx * self._rads_per_pix_x
        dr = np.sqrt(ddx * ddx + ddy * ddy)
        return dr

    # ------------------------------------------------------------------------------
    #           near_field_psf_expfit
    # ------------------------------------------------------------------------------
    def _near_field_psf_expfit(self, modified_angle: np.ndarray, a0: float, k: float, b0: float) -> float:
        """
        Evaluates an exponential function fitted to the "log" of the near field point spread function. The angle used
        in the near field must have been adjusted for the elliptical shape of the near field PSF.

        Parameters
        ----------
        modified_angle: np.ndarray[N]
            An array of modified angle values in radians. The modified angle corrects the Y coordinate of the radial
            distance for elipticity before converting to radians
        a0: float
            Amplitude of the exponential decay of the natural logarithm of the near field PSF.
        k:  float
            Slope of the exponential decay of the natural logarithm of the near field PSF.
        base: float
            Asymptotic value of the exponential decay of the natural logarithm of the near field PSF.

        Returns
        -------
            The value of the natural logairthm of the near field PSF for the given array of modified angles.
        """

        return (a0 * np.exp(-k * modified_angle) + b0)

    # ------------------------------------------------------------------------------
    #           evaluate_near_field_psf
    # ------------------------------------------------------------------------------
    def near_field_psf(self, near_field_coefficients: CentralFieldScatteringCoeffs, psf: np.ndarray):
        """
        Returns the near field psf for every pixel in the supplied PSF array

        :param near_field_coefficients:
        :param psf:
        :return:
        """

        dx, dy = self.dx_and_dy_distances_of_pixels(psf)
        angles = self.get_modified_angle_of_pixels(dx, dy)
        logsignal = self._near_field_psf_expfit(angles, near_field_coefficients.a0, near_field_coefficients.k, near_field_coefficients.b0)
        signal = np.exp(logsignal)
        return signal

    # ------------------------------------------------------------------------------
    #           fit_modified_harvey_curve_to_vis
    # ------------------------------------------------------------------------------
    def fit_near_field_params_to_vis_psf(self, psf_farfield_removed: np.ndarray) -> CentralFieldScatteringCoeffs:
        """
        Fits a modified harvey point spread function to the measured VIS psf.
        This currently only fits in the horizontal direction and is manually adjusted
        to fit to the regions of interest
        :return:
        """

        dx, dy = self.dx_and_dy_distances_of_pixels(psf_farfield_removed)
        angles = self.get_modified_angle_of_pixels(dx, dy)

        good = (angles > 0.00050) & (angles < 0.0050) & (psf_farfield_removed > 0)                                              # Only select points within the valid near field region

        target_base = -30.0                                                                                             # Set the target value for the bas asymptotic log value at wide angle
        dextra = np.zeros([10000]) + target_base                                                                        # and then add 10,000 dummy points that set the asymptote of the fit
        dangle = np.zeros([10000]) + 0.08                                                                               # at the far end of the near out of field. This draws the fit down to the asymptote. There must be a better way but this is good enough
        x = np.hstack((angles[good], dangle))                                                                              # add the dummy angle points to the list of angle values
        y = np.hstack((np.log(psf_farfield_removed[good]), dextra))                                                    # add the dummy asymptote values to the end
        base = target_base                                                                                              # Set up the initial default fo rthe curve fit, this is the asymptote
        a0 = -10 - base                                                                                                 # Set up the initial value for the amplitude of the exponential decay. This seems to work fo rthe moment
        k = 0.69314 / 0.001451                                                                                          # Set up the slope of the exponential decay. Use approximate half height log 2 divided by half height
        initial_params = a0, k, base
        options = {'xtol': 1.49012e-10, 'ftol': 1.49012e-10}
        pnew, pcov = scipy.optimize.curve_fit(self._near_field_psf_expfit, x, y, initial_params, method='lm', **options)
        a0 = pnew[0]
        k = pnew[1]
        base = pnew[2]

        psf_coeffs = CentralFieldScatteringCoeffs(a0=a0, k=k, b0=base, ellipticity_y=self._elliptical_y_factor)

        angle = np.linspace(0.0, 0.080, 10000)
        centraltheory = self._near_field_psf_expfit(angle, a0, k, base)

        logcentralpsf = self._near_field_psf_expfit(angle, a0, k, base)
        nearfield_theory = np.exp(logcentralpsf)

#        farfield_theory = modified_harvey_stray_light_bsdf( angle, far_field_harvey_coeffs)
#        totalpsf_theory = farfield_theory + nearfield_theory

        plt.clf()
        plt.plot(angles[good], np.log(psf_farfield_removed[good]), ',')
        plt.plot(angle, centraltheory, '-')
        plt.grid()

        plt.figure()
        plt.clf()
#        plt.plot( angle, np.log(farfield_theory), '.')
        plt.plot(angle, logcentralpsf, '.')
        plt.grid()

        return psf_coeffs


# ****** AND ANOTHER VERSIONS ****
# ------------------------------------------------------------------------------
#           fit_modified_harvey_curve_to_vis
# ------------------------------------------------------------------------------
def fit_central_field_to_psf(channel: str, psf: np.ndarray, center: Coords) -> HarveyScatteringCoeffs:
    """
    Fits a modified harvey point spread function to the measured VIS psf.
    This currently only fits in the horizontal direction and is manually adjusted
    to fit to the regions of interest
    :return:
    """

    specs = OpticsSpecs(channel)
    rads_per_pix = specs.radians_per_detector_pixel
    r = PSFTools.get_radial_distance_of_pixels(psf, center)
    angle = r * rads_per_pix
    good = (angle > 0.008) & (psf > 0.0)
    x = np.log10(angle[good])
    y = np.log10(psf[good])
    coeffs = np.polyfit(x, y, 1)
    S = -coeffs[0]
    l = 0.001
    bb = coeffs[1] - S * np.log10(l)
    b0 = 10.0**bb
    far_field_harvey_coeffs = HarveyScatteringCoeffs(S=S, l=l, b0=b0)
    farfield_theory = modified_harvey_stray_light_bsdf(angle, far_field_harvey_coeffs)

    psf_farfield_removed = psf - farfield_theory
    dx, dy = get_dx_and_dy_distances_of_pixels(psf, center)
    dy *= rads_per_pix * 0.88586387                         # * psf.shape[1]/psf.shape[0]
    dx *= rads_per_pix

    dy = np.abs(dy)
    dx = np.abs(dx)
    dr = np.sqrt(dx * dx + dy * dy)
    good = (dr > 0.00050) & (dr < 0.0050) & (psf_farfield_removed > 0)
    plt.clf()
    plt.plot(dr[good], np.log(psf_farfield_removed[good]), ',')
    plt.grid()

    target_base = -30
    dextra = np.zeros([10000]) + target_base
    dangle = np.zeros([10000]) + 0.08
    x = np.hstack((dr[good], dangle))
    y = np.hstack((np.log(psf_farfield_removed[good]), dextra))
#    x = dr[good]
#    y = np.log(psf_farfield_removed[good])
    base = target_base
    a0 = -10 - base
    k = 0.69314 / 0.001451  # log 2 divided by half height

    initial_params = a0, k, base
    options = {'xtol': 1.49012e-10, 'ftol': 1.49012e-10}
    pnew, pcov = scipy.optimize.curve_fit(near_field_psf_expfit, x, y, initial_params, method='lm', **options)
    a0 = pnew[0]
    k = pnew[1]
    base = pnew[2]

    angle = np.linspace(0.0, 0.080, 10000)
    centraltheory = near_field_psf_expfit(angle, a0, k, base)
    plt.plot(angle, centraltheory, '-')
    logcentralpsf = near_field_psf_expfit(angle, a0, k, base)
    farfield_theory = modified_harvey_stray_light_bsdf(angle, far_field_harvey_coeffs)
    nearfield_theory = np.exp(logcentralpsf)
    totalpsf_theory = farfield_theory + nearfield_theory

    plt.figure()
    plt.clf()
    plt.plot(angle, np.log(farfield_theory), '.')
    plt.plot(angle, logcentralpsf, '.')
    plt.grid()

    return far_field_harvey_coeffs
