from typing import Tuple, Dict
import numpy as np
import argparse
import matplotlib.pyplot as plt
from altius_l1pp_preflight.configuration import AltiusConfigurationKey, OpticsSpecs, PSFSpecs
from altius_l1pp_preflight.psf import ModifiedHarveyPSF, PSFTools


def main():
    parser = argparse.ArgumentParser(description='Process PSF image and fit harvey straylight coefficients.')
    parser.add_argument('--channel', action='store', required=True, help='Analyze the uv, vis or nir channel')
    parser.add_argument('--wavelength', action='store', required=True, type=float, help='The PSF wavelength. Must match a wavelength entry in the psf.yaml')
    parser.add_argument('--date', action='store', required=True, help='The data of the PSF entry, Typically in the format yyyy-mm-dd. Must match a wavelength entry in the psf.yaml')

    commandline = parser.parse_args()
    channel = commandline.channel.lower()
    wavelen = commandline.wavelength
    datestr = commandline.date
    print('Analyzing channel [{}] at {} nm for {}'.format(channel, wavelen, datestr))
    # channel = 'uv'
    # wavelen = 350.0
    # datestr = '2021-03-15'

    config = AltiusConfigurationKey(channel=channel, wavelength=wavelen, datestr=datestr)
    optics = OpticsSpecs(config)
    psfspecs = PSFSpecs(config)[config.wavelength]
    psf, center = PSFTools.load_psf(config)
    harvey_coeffs = ModifiedHarveyPSF.fit_modified_harvey_curve_to_psf(optics, psfspecs, psf, center, fignum=1)
    print('Harvey Fit for [{}] at {} nm for {}'.format(channel, wavelen, datestr))
    print('S = {:17.15f}'.format(harvey_coeffs.S))
    print('l = {:6.4f}'.format(harvey_coeffs.l))
    print('B0= {:20.15e}'.format(harvey_coeffs.b0))
    plt.show()
    print('Done')