import numpy as np
import sasktran.geometry as geo
import matplotlib.pyplot as plt
from netCDF4 import Dataset
from scipy.signal import convolve2d


class ProfileProvider:
    """Provides a method for computing/interpolating the desired raw
    limb radiance profile onto an image.  Since it's using linear interpolation,
    it requires that the raw data provided extends beyond the limits requested
    by the grid to be interpolated on"""

    def __init__(self, raw_data_points: np.ndarray, altitude_grid_km: np.ndarray):
        self.__raw_data_points: np.ndarray = raw_data_points
        self.__min_alt_km: np.ndarray = altitude_grid_km[0]
        self.__max_alt_km: np.ndarray = altitude_grid_km[-1]
        self.__altitude_grid_km: np.ndarray = altitude_grid_km

    def _set_data_points(self, altitude_grid_km: np.ndarray, raw_data_points: np.ndarray):
        self.__altitude_grid_km = altitude_grid_km
        self.__raw_data_points = raw_data_points

    # Linear 1D interpolation
    def interpolate(self, grid_points: np.ndarray, digitize=False, n_bits=16,
                    saturation_portion=0.9, wavelength_nm=500.0) -> np.ndarray:

        assert (grid_points[0] >= self.__min_alt_km), \
            'Grid point ' + str(grid_points[0]) + ' km is lower than minimum altitude of data ' \
            + str(self.__min_alt_km)

        grid_points = np.sort(grid_points)
        idx = np.argsort(self.__altitude_grid_km)
        xk = self.__altitude_grid_km[idx]
        yk = self.__raw_data_points[idx]
        # Interpolate, filling the high-altitude values (if necessary) with the highest altitude grid value
        y = np.interp(grid_points, xk, yk, right=yk[-1])

        if digitize:
            abs_saturation_val = np.max(y) / saturation_portion
            max_val = saturation_portion * (2**n_bits-1)
            y *= max_val / abs_saturation_val
            y = y.astype(np.dtype('uint' + str(n_bits)))

        return y


class SimpleExponentialProfile(ProfileProvider):
    """Uses the scale height to produce a decaying signal starting at
    1.0 on the bottom of the profile"""

    def __init__(self, scale_height_km: float, scale=1.0, h_bottom_km=0.0, h_top_km=150.0, n_pts=4000, knee_km=3.0):
        grid = np.linspace(h_bottom_km, h_top_km, n_pts)
        y = scale*np.power(10.0, -grid/scale_height_km)
        knee = int((knee_km - h_bottom_km) * n_pts / (h_top_km - h_bottom_km))
        y[0:knee] = y[knee]
        super().__init__(y, grid)


class NetCdfProfile(ProfileProvider):
    """Reads in the NetCdf (*.nc) radiance file provided by Qinetiq used in the Altius SPS
     and interpolates onto the corresponding grid.  Basically a handy way of getting typical limb profiles
     that are wavelength-dependent.  Calculated from MODTRAN.
    Extrapolation on the grid just copies the last value from the data, instead of trying
    to fit a polynomial to it."""

    def __init__(self, nc_file: str, h_bottom_km=0.0, h_top_km=150.0, n_pts=4000):
        grid = np.linspace(h_bottom_km, h_top_km, n_pts)
        self._nc_file = nc_file
        super().__init__(np.zeros(grid.shape), grid)

    def _read_limb_radiance(self, wavelength_nm=500.0) -> (np.ndarray, np.ndarray):
        rootgrp = Dataset(self._nc_file, 'r', format='NETCDF4')

        rad = rootgrp.variables['radiance'][:, :, 0]  # I from stokes vector
        wl = rootgrp.variables['wavelength'][:]
        h = rootgrp.variables['apparant_tp_altitude'][:]  # built-in typo

        # TODO h[km], and wavelength[nm], but can add checks on the units to scale properly
        wlidx = (np.abs(wl - wavelength_nm)).argmin()
        rootgrp.close()
        return h[h > -100], rad[h > -100, wlidx]

    def interpolate(self, grid_points: np.ndarray, digitize=False, n_bits=16,
                    saturation_portion=0.9, wavelength_nm=500.0) -> np.ndarray:
        h, limb = self._read_limb_radiance(wavelength_nm)
        super()._set_data_points(h, limb)
        interp_profile = super().interpolate(grid_points, digitize, n_bits,
                                             saturation_portion, wavelength_nm)
        return interp_profile


class LimbImage:

    def __init__(self, image: np.ndarray, tp_height_grid: np.ndarray):
        self.image: np.array = image
        self.tp_height_grid: np.ndarray = tp_height_grid

    def add_shot_noise(self):
        pass


class LimbImageProvider:
    """Base class to construct a 2D image corresponding to a typical
    limb viewing of the atmosphere from a satellite.  It can be
    constructed in several different ways - such as from OSIRIS
    radiance profiles, or using a simple exponential profile, or
    from ALTIUS SPS 2D image output"""

    def __init__(self, profile_provider: ProfileProvider, bottom_fov_km=0.0,
                 fov_deg=2.0, num_pixels_x=1500, num_pixels_y=1500,
                 earth_curvature=False, wavelength_nm=500):
        self.profile_provider: ProfileProvider = profile_provider
        self.num_pixels_x: int = num_pixels_x
        self.num_pixels_y: int = num_pixels_y
        self.fov_deg: float = fov_deg
        self.image: np.array = np.zeros((num_pixels_y, num_pixels_x))
        self.bottom_fov_km: float = bottom_fov_km
        self.earth_curvature = earth_curvature
        self.grid: np.ndarray = self.determine_altitude_grid(earth_curvature)

    def produce_image(self, digitize=False, n_bits=16, saturation_portion=0.9,
                      wavelength_nm=500.0) -> LimbImage:

        if not self.earth_curvature:
            y = self.profile_provider.interpolate(self.grid[:, 0],
                                                  digitize=digitize, n_bits=n_bits,
                                                  saturation_portion=saturation_portion,
                                                  wavelength_nm=wavelength_nm)
            self.image = np.ones(self.image.shape)*y.reshape((len(y), 1))
        else:
            for ix in range(0, self.num_pixels_x):
                y = self.profile_provider.interpolate(self.grid[:, ix],
                                                      digitize=digitize, n_bits=n_bits,
                                                      saturation_portion=saturation_portion,
                                                      wavelength_nm=wavelength_nm)
                self.image[:, ix] = y

        return self.image

    def determine_altitude_grid(self, spherical=False) -> np.ndarray:
        """Uses approximate R_E = 6378 km and h_sat = 700 km
        to determine vertical grid, assuming 1 arc min == 1 km
        TODO could be changed to do more ray-tracing"""
        grid_2d = np.ones(self.image.shape)
        if spherical is False:
            top_km = self.fov_deg*60 - self.bottom_fov_km
            grid_1d = np.linspace(self.bottom_fov_km, top_km, self.image.shape[0])
            grid_2d = grid_2d * grid_1d.reshape((len(grid_1d), 1))
        else:
            raise NotImplementedError("Have not implemented geoid ray-tracing code")
        return grid_2d

    def add_bright_earth(self):
        pass

    def produce_science_pixel_image(self, kernel: np.ndarray, ds_factor=3, digitize=False, n_bits=14,
                                    saturation_portion=0.9) -> np.ndarray:
        # Kernel must have same length as ds_factor
        img = convolve2d(self.produce_image(digitize, n_bits, saturation_portion),
                         kernel, mode='same')
        return img[::ds_factor, ::ds_factor]
