from typing import Tuple, Dict
import numpy as np
import scipy.signal
from ..configuration import OpticsSpecs, PSFSpecs, WindowTransforms, AltiusConfigurationKey
from ..psf import PSFTools, L1CConvolutionMatrix, NOOFStrayPSF
from .highaltitudeproxy_stray_light_fitter import HighAltitudeProxyStrayLightFitter_FFTConvolved


# ------------------------------------------------------------------------------
#           AltiusStrayLightAlgorithm
# ------------------------------------------------------------------------------
class AltiusStrayLightAlgorithm:

    def __init__(self, config: AltiusConfigurationKey):

        self._config = config                                                                                           # Save the configuratyion key
        self._optics = OpticsSpecs(config)                                                                              # Get the optics specifications
        self._psfspecs = PSFSpecs(self._config)[config.wavelength]                                                  # Get the PSF specifications
        self._straypsf = NOOFStrayPSF(config)
        self._transform = WindowTransforms(self._optics.detector_specs.window_specs)                                    # Make a Window Transform objecty
        self._harvey_coeffs = self._psfspecs.far_field_harvey_coeffs                                                    # Fetch the Modified Harvey PSF for this configuration
        self._stray_light_fitter = HighAltitudeProxyStrayLightFitter_FFTConvolved(config, self._harvey_coeffs, **self._psfspecs.high_altitude_proxy_stray_light_algorithm)          # configure the high altitude proxy fitting algorithm
        matrices = self.load_convolution_matrices()                                                                     # Fetch the convolution matrics
        self._amatrix = matrices[0]                                                                                     # amatrix is the infield convolution matrix
        self._horizmatrix = matrices[1]                                                                                 # horizmatrix in the horizontal convolution matrix
        self._bmatrix = matrices[2]                                                                                     # bmaxtrix is the sum of amatrix and horizmatrix
        self._binv = self._bmatrix.Inverse()
        self._last_infield_stray_estimate: np.ndarray = None
        self._last_horizontal_stray_estimate: np.ndarray = None
        self._last_vertical_stray_estimate: np.ndarray = None

    def set_cache_dir(self, cachedir):
        self._stray_light_fitter.set_cache_dir(cachedir)

    @property
    def infield_stray_estimate(self) -> np.ndarray:
        return self._last_infield_stray_estimate

    @property
    def horizontal_stray_estimate(self) -> np.ndarray:
        return self._last_horizontal_stray_estimate

    @property
    def vertical_stray_estimate(self) -> np.ndarray:
        return self._last_vertical_stray_estimate

    # ------------------------------------------------------------------------------
    #           load_convolution_matrices
    # ------------------------------------------------------------------------------
    def load_convolution_matrices(self) -> Tuple[L1CConvolutionMatrix, L1CConvolutionMatrix, L1CConvolutionMatrix]:

        a = PSFTools.load_convolution_matrix(self._config, matrix_type='infield')
        h = PSFTools.load_convolution_matrix(self._config, matrix_type='horizontal_noof')
        b = a + h
        return a, h, b

    # ------------------------------------------------------------------------------
    #           calculate_infield_stray
    # ------------------------------------------------------------------------------
    def estimate_infield_stray(self, measured_signal: np.ndarray) -> np.ndarray:
        """
        Estimates the in-field stray light by performing the Fourier transform convolution of the Modified Harvey
        PSF with the measured L1C signal. It should really used the true signal but this is pretty darn accurate for
        the ALTIUS VIS channel. This can be replaced with a much faster matrix equation at a later date if we
        generate convolution matrices just using the modified harvey PSF.
        """

        psf_infieldstray = self._straypsf.detector_image()                                                              # Get the stray PSF (no primary lobe) for the detector window
        X = self._transform.l1c_to_detector(measured_signal)                                                            # Get the infield signal, from PSF and horizontal
        infield_stray_full = scipy.signal.fftconvolve(X, psf_infieldstray, mode='same')
        l1c_infield_stray = self._transform.detector_to_l1c(infield_stray_full)
        self._last_infield_stray_estimate = l1c_infield_stray
        return l1c_infield_stray

    # ------------------------------------------------------------------------------
    #           estimate_horizontal_stray
    # ------------------------------------------------------------------------------
    def estimate_horizontal_stray(self, measured_signal: np.ndarray) -> np.ndarray:

        l1c_horiz_stray = self._horizmatrix @ measured_signal
        self._last_horizontal_stray_estimate = l1c_horiz_stray
        return l1c_horiz_stray

    # ------------------------------------------------------------------------------
    #           estimate_horizontal_stray
    # ------------------------------------------------------------------------------
    def estimate_vertical_stray(self, modified_measured_signal: np.ndarray, measured_error: np.ndarray, heightskm: np.ndarray, angles_radians: np.ndarray) -> np.ndarray:

        numcolumns = modified_measured_signal.shape[1]                                                                  # Get the number of columns in this L1C product
        l1c_vertical_stray = np.zeros_like(modified_measured_signal)                                                    # Make an L1C output array

        for icol in range(numcolumns):                                                                                  # For each column in the array
            y = modified_measured_signal[:, icol]                                                                       # Fetch the modified signal
            dy = measured_error[:, icol]                                                                                # Fetch the error bar
            self._stray_light_fitter.fit_stray_light_curve(heightskm, angles_radians, y, dy, icol)                      # do a fit to the raw radiance in the middle of the scene
            l1c_vertical_stray[:, icol] = self._stray_light_fitter.fitted_stray_light(angles_radians, icol)              # and return the fitted stray light profile for all angles/bins/heights
        self._last_vertical_stray_estimate = l1c_vertical_stray
        return l1c_vertical_stray

    # ------------------------------------------------------------------------------
    #           do_analysis
    # ------------------------------------------------------------------------------
    def estimate_true_profile(self,
                              measured_signal: np.ndarray,
                              measured_error: np.ndarray,
                              altitude: np.ndarray,
                              angles: np.ndarray) -> np.ndarray:
        """
        Implements the *High Altitude Proxy* high level algorithm

        Parameters
        ----------
        measured_signal: np.ndarray[Ky,Kx]
            The L1C measured signal.  This is a 2-D array
        measured_error: np.ndarray[Ky, Kx]
            The error on the L1C measured signal
        altitude: np.ndarray[Ky]
            The altitude (in kms?) of the L1C rows
        angles: np.ndarray[Ky]
            The vertical angle in radians of each row from the boresight

        Returns
        -------
        np.ndarray[Ky, Kx]
            The true estimate of the L1C profile with straylight removed.
        """

        true_estimate = measured_signal
        for i in range(2):
            infield_stray = self.estimate_infield_stray(true_estimate)                                                  # Estimate the infield stray light using the measured signal (should use true signal but this is close for vis)
            horiz_stray = self.estimate_horizontal_stray(true_estimate)                                                 # Estimate the horizontal stray light using the measured signal (should use true signal but this is close for vis)
            M = measured_signal - infield_stray - horiz_stray                                                           # Modified measured signal with infield and horizontal removed. ALl that is left is vertical plus true
            vertical_stray = self.estimate_vertical_stray(M, measured_error, altitude, angles)                          # Estimate the vertical stray light
            corrected_signal = measured_signal - vertical_stray                                                         # Remove JUST the vertical stray light estimate from the original measured ste
            true_estimate = self._binv @ corrected_signal                                                              # and do the matrix deconvolution which includes infield and horizontal corrections.

        return true_estimate
