import numpy as np

# ------------------------------------------------------------------------------
#           make_uv_test_signal
# ------------------------------------------------------------------------------
def make_uv_test_signal_sharp_feature(Sy: int, maxvalue=14000.0):
    """
    Creates a simple profile with a sharp square hat feature. The L1C struggles to follow
    this sharp feature and it creates oscillation. It is a simple exponential decay with a scale heightthat matches the
    expected atmospheric signal, ie. one order of magnitude every 15 km, where 15 km is approx 75 science window pixels
    in the UV channel.
    """
    profile = np.zeros([Sy])
    profile[21:50] = maxvalue
    Npts = Sy - 50
    x = -np.arange(Npts) / 75.0
    profile[50:] = maxvalue * np.power(10, x)
    height = np.arange(Sy)*100.0/Sy
    return profile, height


# ------------------------------------------------------------------------------
#           make_uv_test_signal_gauss_feature
# ------------------------------------------------------------------------------
def make_uv_test_signal_gauss_feature(Sy: int, maxvalue=14000.0):

    height = np.arange(Sy)*100.0/Sy
    xg = (height-18.0)/5.0
    g = np.exp( -xg * xg / 2.0 )                # Get a gaussian curve peaking at 18 km and falling off quite quickly
    x = (height - 18.0) / 15.0
    z = np.power( 10.0, -x)                     # Get an exponential decay
    low = height < 18.0                         # and make the expoenntial zero below 18 km
    high = ~low
    z[low] = 0.0
    q = np.zeros_like(g)
    q[low] = g[low]                             # Use the gaussian below 18 km
    q[high] = (g[high] + z[high])/2             # and combine the gaussian and the exponential above 18 km
    q *= maxvalue
    return q, height

# ------------------------------------------------------------------------------
#           make_true_answer
# ------------------------------------------------------------------------------
def make_true_answer(profile, array_window_specs, science_window_only: bool=False) -> np.ndarray:
    """
    Makes a 2-D detector sized window with the true 1-D science profile superimposed at the correct location.
    The true profile can be extended into horizontal near out of field regions.
    """

    NOOFx = array_window_specs[0]
    NOOFy = array_window_specs[1]
    DOFSx = array_window_specs[2]
    DOFSy = array_window_specs[3]
    Dx = array_window_specs[4]
    Dy = array_window_specs[5]
    SOFSx = array_window_specs[6]
    SOFSy = array_window_specs[7]
    Sx = array_window_specs[8]
    Sy = array_window_specs[9]

    image = np.zeros([Dy, Dx])                                                  # Create a window the size of the detector window
    if (science_window_only):                                                   # if on the science window region is required
        for i in range(Sx):                                                     # then for each column of the science window
            image[SOFSy:SOFSy + Sy, SOFSx + i] = profile                        # insert the 1-D profile
    else:                                                                       # otherwise we also want stray light
        leftnoof = -DOFSx                                                       # Get the left edge of the horizontal stray light region in detector window coordinates
        rightnoof = leftnoof + NOOFx                                            # Get the right edge of the horizontal stray light in detector window corodinates
        if leftnoof < 0:                                                        # make sure the left edge is within range
            leftnoof = 0
        if rightnoof > Dx:                                                      # make sure the right edge is within range
            rightnoof = Dx
        for i in range(leftnoof, rightnoof):                                    # and insert the 1-D profile
            image[SOFSy:SOFSy + Sy, i] = profile                                # into all columns of the detector window
    return image
