from typing import Tuple, List
import os.path
import numpy as np
import pickle
import scipy.io
import scipy.signal
import scipy.interpolate
import matplotlib.pyplot as plt
from l1c_deconvolution_matrix.deconvolution_matrix import generate_convolution_matrix, generate_horizontal_stray_light_matrix, WINDOW_DETECTOR
from l1c_deconvolution_matrix.matrix_file_io import write_to_netcdf, load_from_netcdf


# ------------------------------------------------------------------------------
#           shift_central_area
# ------------------------------------------------------------------------------
def shift_central_area( psf, y0, x0, deltay, deltax, L: int=6):

    L = 6
    samps_per_pix = 100
    N = L * samps_per_pix
    iy0 = int(y0) - L//2
    iy1 = iy0 + L
    ix0 = int(x0) - L//2
    ix1 = ix0 + L
    data =psf[iy0:iy1, ix0:ix1]
    base = np.min(data)
    data = data - base
    highres = np.zeros([N, N])

    for iy in range(L):
        for ix in range(L):
            ly0 = samps_per_pix * iy
            lx0 = samps_per_pix * ix
            highres[ly0:ly0 + samps_per_pix, lx0:lx0 + samps_per_pix] = data[iy,ix]

    yshift = int(deltay * samps_per_pix + 0.5)
    xshift = int(deltax * samps_per_pix + 0.5)
    highshift = np.roll(highres, (yshift,xshift), axis=(0,1))
    lowshift = np.sum( highshift.reshape([L,samps_per_pix, L, samps_per_pix]), axis=(1,3)) / (samps_per_pix * samps_per_pix)
    lowshift = lowshift + base
    return lowshift, base


# ------------------------------------------------------------------------------
#           shift_psf_to_image_centre
# ------------------------------------------------------------------------------
def shift_psf_to_image_centre(psf: np.ndarray, y0: float, x0: float) -> Tuple[np.ndarray, float, float]:
    """
    Shifts the PSF so the center of the PSF is properly centered for the FFT convolution. This means placing
    the center of the PSF at Y = 999.5 and X at 751.5 for the detector window

    Most of the image is shifted with scipy.interpolate.interpn but special care is taken with teh narrow central peak.
    The pixels in the vicinity of the narrow central peak are subdivided into 1/100 pixel areas. These pixels are then
    shifted to the origin (typiucally 0.5 pixels)
    """

    Dy, Dx = psf.shape
    logpsf = np.log( psf )
    bad = logpsf < -30.0
    logpsf[bad] = -30.0
    newx0 = float(Dx // 2 - 1) + 0.5
    newy0 = float(Dy // 2 - 1) + 0.5
    deltax = (newx0 - x0)
    deltay = (newy0 - y0)

    assert( (np.abs(deltay) < 1.1) and (np.abs(deltax) < 1.1) )                 # We can only do small shifts for the moment. Update the code for bigger shifts
    L = 6                                                                       # Look at a 6 x 6 pixel region around the center.
    lowshift, base = shift_central_area( psf, y0, x0, deltay, deltax, L=L)      # and shift the central region
    iy0 = int(y0) - L//2
    iy1 = iy0 + L
    ix0 = int(x0) - L//2
    ix1 = ix0 + L

    x = np.arange(Dx)
    y = np.arange(Dy)
    actualx = x - deltax
    actualy = y - deltay
    ax = np.tile(actualx.reshape([1, Dx]), [Dy, 1])
    ay = np.tile(actualy.reshape([Dy, 1]), [1, Dx])
    locations = np.zeros([Dy, Dx, 2])
    locations[:,:,1] = ax
    locations[:,:,0] = ay
    points = (y, x)
    newpsf = scipy.interpolate.interpn( points, psf, locations, bounds_error=False, fill_value=1.0E-15)
    newpsf[iy0:iy1, ix0:ix1]=lowshift
    return newpsf, newy0, newx0

# ------------------------------------------------------------------------------
#           load_psf_cell_array
# ------------------------------------------------------------------------------
def load_psf_cell_array() -> List[np.ndarray]:
    """
    Loads in the UV psf which is stored with this test procedure. This UV image has two extra dots, one either side of
    the central core. It was released by OIP in the summer of 2020.
    """
    basedir, name = os.path.split(__file__)
    psfdata = scipy.io.loadmat(os.path.join(basedir, 'psf_uv_300nm_EOL.mat'))
    psfarray = np.array(psfdata['psf_uv'], dtype='float64')
    psfarray = psfarray / np.sum(psfarray)
    cellpsf = [psfarray]                                                              # Create the cell array
    return cellpsf


# ------------------------------------------------------------------------------
#           make_uv_window_specs
# ------------------------------------------------------------------------------
def make_uv_window_specs():
    """
    Makes the windowing specs for thgis test. This test makes a relatively large number of L1C horizontal pixels (KBINx=5, Kx=34)
    as that provides a bigger strain on the system and allows for more detailed analysis during debugging sessions. It is possible
    to use ( KBINx=2, Kx=85) although this uses a large amount of memory (14 GB) and may be too large for many test systems.
    """

    NOOFx = 846                             # X Number of pixels in Near-Out-Of-Field window. Specified in detector pixels.
    NOOFy = 846                             # Y Number of pixels in Near-Out-Of-Field window. Specified in detector pixels
    DOFSx = -329                            # X Offset of detector window in the near out-of-field window. Specified in NOOF pixels
    DOFSy = -577                            # Y Offset of detector window in the near out-of-field window, Specifie di NOOF pixels.
    Dx = 1504                               # X number of pixels in detector window
    Dy = 2000                               # Y number of pxiels in detector window

    SOFSx = 494                             # X Offset of science window in detector pixels
    SOFSy = 742                             # Y Offset of science window in detector pixels
    Sx = 510                                # X number of pixels in science window. Specified in detector pixels.
    Sy = 513                                # Y number of pixels in science window. Specified in detector pixels.

    LBINx = 3                               # X Number of science window pixels binned per L1B pixel.
    LBINy = 3                               # Y Number of science window pixels binned per L1B pixel.
    Lx = 170                                # X number of L1B pixels. Note: Sx must equal Lx * LBINx
    Ly = 171                                # Y number of L1B pixels. Note: Sy must equal Ly * LBINy

    KBINx = 10                              # X Number of L1B window pixels binned per L1C window pixel.
    KBINy = 1                               # Y Number of L1B window pixels binned per L1C window pixel.
    Kx = 17                                 # X Number of L1C pixels. Note: Lx must equal Kx * KBINx
    Ky = 171                                # Y Number of L1C pixels. Note: Ly must equal Ky * KBINy

    # LBINx = 1                               # X Number of science window pixels binned per L1B pixel.
    # LBINy = 1                               # Y Number of science window pixels binned per L1B pixel.
    # Lx = 510                                # X number of L1B pixels. Note: Sx must equal Lx * LBINx
    # Ly = 513                                # Y number of L1B pixels. Note: Sy must equal Ly * LBINy
    #
    # KBINx = 30                              # X Number of L1B window pixels binned per L1C window pixel.
    # KBINy = 1                               # Y Number of L1B window pixels binned per L1C window pixel.
    # Kx = 17                                 # X Number of L1C pixels. Note: Lx must equal Kx * KBINx
    # Ky = 513                                # Y Number of L1C pixels. Note: Ly must equal Ky * KBINy

    array_window_specs = [NOOFx, NOOFy,
                          DOFSx, DOFSy, Dx, Dy,
                          SOFSx, SOFSy, Sx, Sy,
                          LBINx, LBINy, Lx, Ly,
                          KBINx, KBINy, Kx, Ky]
    return array_window_specs


# ------------------------------------------------------------------------------
#           make_uv_test_signal
# ------------------------------------------------------------------------------
def make_uv_test_signal_sharp_feature(Sy: int, maxvalue=14000.0):
    """
    Creates a simple profile with a sharp square hat feature. The L1C struggles to follow
    this sharp feature and it creates oscillation. It is a simple exponential decay with a scale heightthat matches the
    expected atmospheric signal, ie. one order of magnitude every 15 km, where 15 km is approx 75 science window pixels
    in the UV channel.
    """
    profile = np.zeros([Sy])
    profile[21:50] = maxvalue
    Npts = Sy - 50
    x = -np.arange(Npts) / 75.0
    profile[50:] = maxvalue * np.power(10, x)
    height = np.arange(Sy)*100.0/Sy
    return profile, height


# ------------------------------------------------------------------------------
#           make_uv_test_signal_gauss_feature
# ------------------------------------------------------------------------------
def make_uv_test_signal_gauss_feature(Sy: int, maxvalue=14000.0):

    height = np.arange(Sy)*100.0/Sy
    xg = (height-18.0)/5.0
    g = np.exp( -xg * xg / 2.0 )                # Get a gaussian curve peaking at 18 km and falling off quite quickly
    x = (height - 18.0) / 15.0
    z = np.power( 10.0, -x)                     # Get an exponential decay
    low = height < 18.0                         # and make the expoenntial zero below 18 km
    high = ~low
    z[low] = 0.0
    q = np.zeros_like(g)
    q[low] = g[low]                             # Use the gaussian below 18 km
    q[high] = (g[high] + z[high])/2             # and combine the gaussian and the exponential above 18 km
    q *= maxvalue
    return q, height
# ------------------------------------------------------------------------------
#           make_true_answer
# ------------------------------------------------------------------------------
def make_true_answer(profile, array_window_specs, science_window_only: bool=False) -> np.ndarray:
    """
    Makes a 2-D detector sized window with the true 1-D science profile superimposed at the correct location.
    The true profile can be extended into horizontal near out of field regions.
    """

    NOOFx = array_window_specs[0]
    NOOFy = array_window_specs[1]
    DOFSx = array_window_specs[2]
    DOFSy = array_window_specs[3]
    Dx = array_window_specs[4]
    Dy = array_window_specs[5]
    SOFSx = array_window_specs[6]
    SOFSy = array_window_specs[7]
    Sx = array_window_specs[8]
    Sy = array_window_specs[9]

    image = np.zeros([Dy, Dx])                                                  # Create a window the size of the detector window
    if (science_window_only):                                                   # if on the science window region is required
        for i in range(Sx):                                                     # then for each column of the science window
            image[SOFSy:SOFSy + Sy, SOFSx + i] = profile                        # insert the 1-D profile
    else:                                                                       # otherwise we also want stray light
        leftnoof = -DOFSx                                                       # Get the left edge of the horizontal stray light region in detector window coordinates
        rightnoof = leftnoof + NOOFx                                            # Get the right edge of the horizontal stray light in detector window corodinates
        if leftnoof < 0:                                                        # make sure the left edge is within range
            leftnoof = 0
        if rightnoof > Dx:                                                      # make sure the right edge is within range
            rightnoof = Dx
        for i in range(leftnoof, rightnoof):                                    # and insert the 1-D profile
            image[SOFSy:SOFSy + Sy, i] = profile                                # into all columns of the detector window
    return image


# ------------------------------------------------------------------------------
#           make_l1c_from_detector_window_image
# ------------------------------------------------------------------------------
def make_l1c_from_detector_window_image(image, array_window_specs):
    """
    MAkes an L1C product given a detector window and a set of specs.
    """

    SOFSx = array_window_specs[6]
    SOFSy = array_window_specs[7]
    Sx = array_window_specs[8]
    Sy = array_window_specs[9]
    LBINx = array_window_specs[10]
    LBINy = array_window_specs[11]
    KBINx = array_window_specs[14]
    KBINy = array_window_specs[15]
    Kx = array_window_specs[16]
    Ky = array_window_specs[17]
    newimage = image[SOFSy:SOFSy + Sy, SOFSx:SOFSx + Sx]                    # Make a new image that collects just the science window from the detector window
    l1b = newimage.reshape([Ky, LBINy * KBINy, Kx, LBINx * KBINx])          # Reshape Y axis into two components and X axis into two components
    l1c = np.sum(l1b, axis=(1, 3))                                          # Sum over the axes to leave the L1C product.
    return l1c                                                              # and return the L1C product


# ------------------------------------------------------------------------------
#           L1C_demo_UV
# ------------------------------------------------------------------------------
def make_horizontal_stray_light_array(cellpsf: List[np.ndarray], array_window_specs, y0=1000.5, x0=752.5):
    """
    This is an example that shows how to make the horizontal stray light convolution matrix. Vignetting is set to an
    empty array which is detected by the python interface and a default of 1.0 is used for vignetting
    """

    NOOFx = array_window_specs[0]
    NOOFy = array_window_specs[1]
    locn = np.array([y0, x0, WINDOW_DETECTOR]).reshape([1, 3])                    # set up the (Y,X) coordinates of the center of the PSF, Y then X
    coeffs = np.ones([1, NOOFy, NOOFx])                                                 # Set all the interpolation coefficients to 1.0
    indices = np.zeros([1, NOOFy, NOOFx])                                               # Set all the PSF indices to zero.
    vignetting = np.zeros([0, 0])
    horiz_stray = generate_horizontal_stray_light_matrix(array_window_specs, vignetting, cellpsf, locn, coeffs, indices)
    return horiz_stray


# ------------------------------------------------------------------------------
#           L1C_demo_UV
# ------------------------------------------------------------------------------
def make_uv_in_field_matrix(cellpsf: List[np.ndarray], array_window_specs, y0=1000.5, x0=752.5):
    """
    This is an example that shows how to make the in-field convolution matrix fo rthe L1C product. This only uses 1
    psf array so it creates an interpolation array (of zeros) that indices the first psf. It also makes an array of
    interpolation coefficients for each psf array. The interpolation array is set to 1.0 everywhere.
    """

    Dx = array_window_specs[4]
    Dy = array_window_specs[5]
    locn = np.array([y0, x0, WINDOW_DETECTOR]).reshape([1, 3])
    coeffs = np.ones([1, Dy, Dx])                                               # Set all the interpolation coefficients to 1.0
    indices = np.zeros([1, Dy, Dx])                                             # Set all the PSF indices to zero.
    a_matrix = generate_convolution_matrix(array_window_specs, cellpsf, locn, coeffs, indices)
    return a_matrix


# ------------------------------------------------------------------------------
#           l1c_matrix_mul
# ------------------------------------------------------------------------------
def l1c_matrix_mul(l1c: np.ndarray, convolution_matrix: np.ndarray) -> np.ndarray:
    """
    Multiplies the convolution matrix by the L1C array. The main ussue here is that L1C comes in as a 2-D image array
    that must be flattened to match the convolution matrix. After multiplication it is un-flattened back to an L1 shape.

    The deconvolution matrix has a column entry and a row entry for every pixel in the L1C arra and indicates the contribution
    of every L1C pixel to every other L1C pixel. You must know that the columns and rows of the deconvolution matrix are
    ordered in X groups of contiguous Y, which is a hangover from Matlab that uses a column major format.

    Parameters
    ----------
    l1c : np.ndarray[Ky, Kx]
        A 2-D array of L1C data product. The first dimension, Ky,  is the Y (vertical) axis and the second, Kx, is the X axis.

    l1c : np.ndarray[Ky*Kx, Ky*Kx]
        A square 2-D convolution (or deconvolution) array/matrix. The size of each dimension must equal Kx * Ky. The first dimension
        indexes rows (Y) of the matrix  an dthe second indexes columns (X).

    Returns
    -------
    np.ndarray[Ky, Kx]
        Returns an array the same shape as `l1c` given by the matrix equation :math:`Y=AX`.
    """
    s = l1c.shape
    Ky = s[0]
    Kx = s[1]
    tt = l1c.transpose().flatten()
    qq = convolution_matrix @ tt
    yf = qq.reshape(Kx, Ky).transpose()
    return yf


# ------------------------------------------------------------------------------
#           uv_test_analysis
# ------------------------------------------------------------------------------
def uv_test_analysis():
    """
    Runs a test that checks out the deconvolution algorithm. This builds the infield and horizontal stray light
    deconvolution matrices and test data. It then builds the inverse matrices and solves the inversion.
    """

    array_window_specs = make_uv_window_specs()                                                                         # Create window specs for the UV region
    SOFSx = array_window_specs[6]                                                                                       # X Offset of science window in detector window
    SOFSy = array_window_specs[7]                                                                                       # Y Offset of science window in detector window
    Sx = array_window_specs[8]                                                                                          # X Size of science window
    Sy = array_window_specs[9]                                                                                          # Y Size of science window
    LBINx = array_window_specs[10]
    LBINy = array_window_specs[11]
    KBINx = array_window_specs[14]
    KBINy = array_window_specs[15]
    Kx = array_window_specs[16]
    Ky = array_window_specs[17]

    profile, height = make_uv_test_signal_gauss_feature(Sy)                                                             # Make a 1-D theoretical true profile. Its a simple exponential decay
    true_image = make_true_answer(profile, array_window_specs)                                                          # Make a true image that spans the entire noof window including horizontal stray light
    true_outoffield = np.array(true_image)                                                                              # Make a true image
    true_outoffield[:, SOFSx:SOFSx + Sx] = 0                                                                            # that only has the horizontal out-of-field

    cellpsf = load_psf_cell_array()                                                                                     # Load in the list of 1 PSF arrays
    y0 = 1000.0                                                                                                         # specify the center of the PSF array, note that centers of pixel are at 0.5 values
    x0 = 752.0                                                                                                          # and
    psf = cellpsf[0]                                                                                                    # get the psf array from the first element.

    newpsf, y0, x0 = shift_psf_to_image_centre(psf, y0, x0)                                                             # Shift the PSF so it is properly centered (as returned in y0 and x0)
    cellpsf = [newpsf]
    psf = newpsf

    # psf[999:1001, 750:754] = psf[998, 750]          # Knock out the central peak.
    # psf[0:SOFSy,:] = 0                              # Take out the PSF in the vertical near out-of-field
    # psf[SOFSy+Sy:, :] = 0                           # Both below and above the science window
    # psf[0:900, 0:SOFSx] = 0                         # take out most of the left edge apart from the bright dots
    # psf[0:900, SOFSx+Sx:] = 0                       # take out most of the left edge apart from the bright dots
    # psf[1100:, 0:SOFSx] = 0                         # take out most of the left edge apart from the bright dots
    # psf[1100:, SOFSx+Sx:] = 0                       # take out most of the left edge apart from the bright dots

    measured_signal = scipy.signal.fftconvolve(true_image, psf, mode='same')                                            # Calculate the measured signal using a convolution
    measured_signal[0:SOFSy, :] = 0                                                                                     # wipe out the lower
    measured_signal[SOFSy + Sy:, :] = 0                                                                                 # and upper parts of the image (these can be left but are never used)

    true_outoffield_stray = scipy.signal.fftconvolve(true_outoffield, psf, mode='same')                                 # Calculate how much stray light scatters in the science window
    true_outoffield_stray[0:SOFSy, :] = 0
    true_outoffield_stray[SOFSy + Sy:, :] = 0

    matrixname = 'matrices_{:1d}_{:1d}.pickle'.format(Ky, Kx)
    if (os.path.exists(matrixname)):
        f = open(matrixname, 'rb')
        a_matrix, h_matrix, Ainv, Binv = pickle.load(f)
        b_matrix = a_matrix + h_matrix
        f.close()
    else:
        h_matrix = make_horizontal_stray_light_array(cellpsf, array_window_specs, y0=y0, x0=x0)                     # make the horizontal stray light convolution matrix
        a_matrix = make_uv_in_field_matrix(cellpsf, array_window_specs, y0=y0, x0=x0)                               # make the in-field stray light convolution matrix
        b_matrix = a_matrix + h_matrix                                                                              # add the two matrices together to get the corrected in-field stray light convolution matrix.
        Binv = np.linalg.inv(b_matrix)                                                                              # Get the deconvolution matrix of the corrected in-field stray light by taking the inverse
        Ainv = np.linalg.inv(a_matrix)
        f = open(matrixname, 'wb')                                                                           # save the data to file, for convenience when debugging
        pickle.dump((a_matrix, h_matrix, Ainv, Binv), f)
        f.close()

    measured_l1c = make_l1c_from_detector_window_image(measured_signal, array_window_specs)
    true_l1c = make_l1c_from_detector_window_image(true_image, array_window_specs)                                # make the true true answer. This is what we want to retrieve.
    true_stray_l1c = make_l1c_from_detector_window_image(true_outoffield_stray, array_window_specs)                 # make the true stray light scattered

    stray_matrix = l1c_matrix_mul(true_l1c, h_matrix)
    infield_matrix = l1c_matrix_mul(true_l1c, a_matrix)
    answer = l1c_matrix_mul(measured_l1c, Binv)

    ZBINy = LBINy * KBINy
    ZBINx = LBINx * KBINx
    idx = Kx//2
    h = height.reshape( [Sy//ZBINy,ZBINy])[:,ZBINy//2]
    yt = true_l1c[:, idx]
    ym = measured_l1c[:, idx]
    ya = answer[:, idx]
    yin = infield_matrix[:, idx]
    yoof = stray_matrix[:, idx]
    diff = np.abs((yt - ya) / yt)

    plt.figure(1)
    plt.clf()
    plt.plot(np.log10(yt), h, 'k-', linewidth=0.5, label='True Profile')
    plt.plot(np.log10(ym), h, 'b-', linewidth=0.5, label='Measured Profile')
    plt.plot(np.log10(ya), h, 'r-', linewidth=0.5, label='Retrieved Profile')
    plt.plot(np.log10(yin), h, 'm-', linewidth=0.5, label='Infield signal')
    plt.plot(np.log10(yoof), h, 'g-', linewidth=0.5, label='Horiz Noof signal')
    plt.xlabel('Log10 Signal')
    plt.ylabel('Nominal height km')
    plt.title('Infield + Horiz Stray Light. {} heights'.format(Ky))
    plt.legend()

    plt.figure(2)
    plt.clf()
    plt.plot(diff*100.0, h)
    plt.xlabel('Percentage difference')
    plt.ylabel('Nominal height km')
    plt.title('Inversion Performance. {} heights'.format(Ky))

    plt.show(block=False)
    print('Done')


if __name__ == '__main__':
    uv_test_analysis()
