"""
This is a study for stray light removal using radiance profiles provided by BISA. This study was initiated in February/March
2021 in preparation for the ALTIUS PKP#3 in March 2021. The study attempts to shed light on the accuracy and efficacy of
the infield, horizontal and vertical stray light removal algorithms. This study was created in response to new data/
information regarding the near out-of-field performance of the ALTIUS optics designed by OIP. Simply put, there was significant
contribution to the measured signal, when compared to ALTIUS system requiremenets at least, from near out of field components.
The near out-of field is looslely defined as the angular region just outside the imaging region of the optics which is not
completely eliminated by the baflle system.

The infield stray light is all light scattered by the point spread function  from one location within the field of view
to other locations within the field of view. This term is handled by the l1c convolution matrix (``a_matrix``) which is akin to a standard
deconvolution algorithm.

The horizontal stray light is due to stray light from horizontal regions adjacent to the left and right of the field of view.
This term is modelled by assuming the horizontal radiance signal is identical (or a linear combination) to the infield signal.
This study simply assumes this statement is true and does not investigate the validity of the statement. This allows horizontal
light to treated as a special case of in-field stray as it is a linear combination of the infield components. The l1c convolution
C++ software is used to create a specialized horizontal convolution matrix (``h_matrix``) specifically for the removal of
horizontal stray light.

The vertical stray light is due to stray light from vertical regions below the field of view. Technically stray light could
also come from above but this is explicitly assumed to be zero. Vertical stray light is removed using the high altitude proxy
algorithm: deviation of the high altitude signal from the expected background signal is attributed to vertical stray light.

The high altitude proxy algorithm must remove high altitude stray light contributions from the infield and horizontal stray
light before fitting vertical stray light. The infield stray light requires convolution of the infield signal with the point
spread function. Ideally this would be a C++ matrix multiplication but is currently implemented as a relatively slow FFTconvolution.
The required C++ matrix would be identical to the standard convolution matrix (``a_matrix``) except it would be generated with the
primary lobe knocked out so it only estimates stray light distribution.

The high altitude proxy uses an iterative procedure where it starts with the measured signal as a good estimate of the "true" signal.
This initial estimate can be used to calculate estimates of horizontal, infield and vertical stray light. These allow new estimates
of the "true" signal to be estimated and the process repeated. Generally only a couple of iterations are required.

The current high altitude proxy algorithm assumes the entire near-out-of-field region can be approximated with a mean, constant value.
This allows fairly rapid estimation of convolution integrals. This is also driven by the fact that the high altitude proxy is a very smoothed
function of the true out-of-field signal and does not contain much information, i.e. there is enough to fit one parameter and there "may" be
enough to fit two if you are lucky (eg mean and gradient).
"""

from typing import Tuple
import os.path
import numpy as np
import xarray as xr
import appdirs
import argparse
import matplotlib.pyplot as plt
from ... configuration import AltiusConfigurationKey, OpticsSpecs
from ... stray_light import AltiusStrayLightAlgorithm, AltiusStrayLightAnalysis_MeasurementSimulator, AltiusStrayLightSystemRequirements
from ... radiance_simulator.bisa_stray_light_stimulus_data.noof_simulus_reader import SimulationRadianceDataBISA


# ------------------------------------------------------------------------------
#           Class StrayLightRemovalStudyBISAProfiles
# ------------------------------------------------------------------------------
class StrayLightRemovalStudyBISAProfiles:

    def __init__(self, wavelength_index: int, retrieval_config: AltiusConfigurationKey, generation_config: AltiusConfigurationKey, simulation: SimulationRadianceDataBISA, cache_dir: str=None, results_dir='./'):

        self._retrieval_config = retrieval_config
        self._generation_config = generation_config
        self._optics = OpticsSpecs(self._retrieval_config)
        self._simulation = simulation
        self._algorithm = AltiusStrayLightAlgorithm(self._retrieval_config)
        self._analysis_simulator = AltiusStrayLightAnalysis_MeasurementSimulator(self._generation_config, simulation)
        self._wavelength = simulation.wavelength[wavelength_index]
        self._wavelength_index = wavelength_index
        self._srd = AltiusStrayLightSystemRequirements()
        self._results_dir = results_dir
        self._allocate_arrays()
        if cache_dir is not None:
            self.set_cache_dir(cache_dir)

    def set_cache_dir(self, cachedir):
        self._analysis_simulator.set_cache_dir(cachedir)
        self._algorithm.set_cache_dir(cachedir)

    # ------------------------------------------------------------------------------
    #           _allocate_arrays
    # ------------------------------------------------------------------------------
    def _allocate_arrays(self):
        N = self._simulation.num_cases
        Kx = self._optics.detector_specs.window_specs.Kx
        Ky = self._optics.detector_specs.window_specs.Ky
        Ty = self._simulation.angle.size

        self._l1c_measured_signal = np.zeros([N, Ky, Kx])
        self._l1c_measured_error = np.zeros([N, Ky, Kx])
        self._l1c_true_signal = np.zeros([N, Ky, Kx])
        self._l1c_true_estimate = np.zeros([N, Ky, Kx])
        self._l1c_true_infield_stray = np.zeros([N, Ky, Kx])
        self._l1c_estimate_infield_stray = np.zeros([N, Ky, Kx])
        self._l1c_true_horizontal_stray = np.zeros([N, Ky, Kx])
        self._l1c_estimate_horizontal_stray = np.zeros([N, Ky, Kx])
        self._l1c_true_vertical_stray = np.zeros([N, Ky, Kx])
        self._l1c_estimate_vertical_stray = np.zeros([N, Ky, Kx])
        self._l1c_altitude = np.zeros([N, Ky])
        self._l1c_angle = np.zeros([N, Ky])
        self._true_signal = np.zeros([N, Ty])
        self._true_altitude = np.zeros([N, Ty])
        self._true_angle = np.zeros([N, Ty])

    # ------------------------------------------------------------------------------
    #           write_to_netcdf
    # ------------------------------------------------------------------------------
    # def write_to_netcdf(self):
    #     l1c_measured_signal               = rootgrp.createVariable ('l1c_measured_signal',           np.float64, ('scene', 'wavelength', 'Ky', 'Kx'))       # noqa
    #     l1c_measured_error                = rootgrp.createVariable ('l1c_measured_error',            np.float64, ('scene', 'wavelength', 'Ky', 'Kx'))       # noqa
    #     l1c_smooth_signal                 = rootgrp.createVariable ('l1c_smooth_signal',             np.float64, ('scene', 'wavelength', 'Ky', 'Kx'))       # noqa
    #     l1c_true_estimate                 = rootgrp.createVariable ('l1c_true_estimate',             np.float64, ('scene', 'wavelength', 'Ky', 'Kx'))       # noqa
    #     l1c_true_infield_stray            = rootgrp.createVariable ('l1c_true_infield_stray',        np.float64, ('scene', 'wavelength', 'Ky', 'Kx'))       # noqa
    #     l1c_estimate_infield_stray        = rootgrp.createVariable ('l1c_estimate_infield_stray',    np.float64, ('scene', 'wavelength', 'Ky', 'Kx'))       # noqa
    #     l1c_true_horizontal_stray         = rootgrp.createVariable ('l1c_true_horizontal_stray',     np.float64, ('scene', 'wavelength', 'Ky', 'Kx'))       # noqa
    #     l1c_estimate_horizontal_stray     = rootgrp.createVariable ('l1c_estimate_horizontal_stray', np.float64, ('scene', 'wavelength', 'Ky', 'Kx'))       # noqa
    #     self._l1c_true_vertical_stray     = rootgrp.createVariable ('l1c_true_vertical_stray',       np.float64, ('scene', 'wavelength', 'Ky', 'Kx'))       # noqa
    #     self._l1c_estimate_vertical_stray = rootgrp.createVariable ('l1c_estimate_vertical_stray',   np.float64, ('scene', 'wavelength', 'Ky', 'Kx'))       # noqa
    #     self._l1c_altitude                = rootgrp.createVariable ('l1c_altitude',                  np.float64, ('scene', 'Ky'))                           # noqa
    #     self._l1c_angle                   = rootgrp.createVariable ('l1c_angle',                     np.float64, ('scene', 'Ky'))                           # noqa
    #     self._true_signal                 = rootgrp.createVariable ('true_signal',                   np.float64, ('scene', 'wavelength', 'Ty'))             # noqa
    #     self._true_altitude               = rootgrp.createVariable ('true_altitude',                 np.float64, ('scene', 'Ty'))                           # noqa
    #     self._true_angle                  = rootgrp.createVariable ('true_angle',                    np.float64, ('scene', 'Ky'))                           # noqa
    #     self._wavelength                  = rootgrp.createVariable ('wavelength',                    np.float64, ('wavelength',))                           # noqa

    # ------------------------------------------------------------------------------
    #           calculate_noise
    # ------------------------------------------------------------------------------
    def calculate_noisy_signal(self, smooth_measured_signal: np.ndarray, sample_noisy_profile: bool) -> Tuple[np.ndarray, np.ndarray]:

        binx = self._optics.detector_specs.window_specs.LBINx * self._optics.detector_specs.window_specs.KBINx
        biny = self._optics.detector_specs.window_specs.LBINy * self._optics.detector_specs.window_specs.KBINy
        binning = binx * biny
        measured_error = np.sqrt(smooth_measured_signal + binning * self._optics.detector_specs.read_out_noise)
        if (sample_noisy_profile):
            rng = np.random.default_rng()                                                                               # Initialize the random number generator
            measured_signal = rng.normal(smooth_measured_signal, measured_error)                                        # Get the noisy total signal.
            bad = measured_signal < 0.0
            measured_signal[bad] = 0.0
            measured_error = np.sqrt(measured_signal)
        else:
            measured_signal = smooth_measured_signal
        return measured_signal, measured_error

    # ------------------------------------------------------------------------------
    #           make_figures
    # ------------------------------------------------------------------------------
    def make_figures(self, make_noisy_profile, stimulus: int):

        figure_folder = 'shotnoise' if make_noisy_profile else 'ideal'
        basedir = os.path.normpath(os.path.join(self._results_dir, figure_folder, 'figures'))
        os.makedirs(basedir, exist_ok=True)

        plt.figure(4, figsize=(10, 7))
        plt.clf()
        icol = self._l1c_measured_signal.shape[-1] // 2
        alt = self._l1c_altitude[stimulus, :]
        plt.plot(np.log10(self._l1c_measured_signal[stimulus, :, icol]), alt, 'g-', label='Measured signal')
        plt.plot(np.log10(self._l1c_true_signal[stimulus, :, icol]), alt, 'k-', label='True signal')
        plt.plot(np.log10(self._l1c_true_estimate[stimulus, :, icol]), alt, 'r-.', label='Estimated signal')
        plt.plot(np.log10(self._l1c_true_infield_stray[stimulus, :, icol]), alt, 'y-', label='True Infield Stray')
        plt.plot(np.log10(self._l1c_estimate_infield_stray[stimulus, :, icol]), alt, 'y-.', label='Estimated Infield Stray')
        plt.plot(np.log10(self._l1c_true_horizontal_stray[stimulus, :, icol]), alt, 'm-', label='True Horizontal Stray')
        plt.plot(np.log10(self._l1c_estimate_horizontal_stray[stimulus, :, icol]), alt, 'm-.', label='Estimated Horizontal Stray')
        plt.plot(np.log10(self._l1c_true_vertical_stray[stimulus, :, icol]), alt, 'b-', label='True Vertical Stray')
        plt.plot(np.log10(self._l1c_estimate_vertical_stray[stimulus, :, icol]), alt, 'b-.', label='Estimated Vertical Stray')
        plt.legend()
        plt.xlabel('Log10(Signal)')
        plt.ylabel('Height Km')
        plt.title('Stray Light Correction for {} nm, scene {}'.format(self._wavelength, stimulus))
        plt.savefig(os.path.join(basedir, 'stray_light_correction_{}_{:1d}_{:1d}.png'.format(figure_folder, int(self._wavelength), stimulus)))

        plt.figure(5)
        plt.clf()
        diff = np.abs(self._l1c_true_estimate[stimulus, :, :] - self._l1c_true_signal[stimulus, :, :]) / self._l1c_true_signal[stimulus, :, :]
        linetype = 'ks' if make_noisy_profile else 'ks-'

        good = (alt <= 50.0) & (alt >= 5.0)
        h, srd = self._srd.srd(self._wavelength)
        maxdev = np.max(diff[good, icol])
        if (maxdev > 1.0):
            maxdev = 1.0
        maxsrd = np.max(srd)
        if (maxsrd > maxdev):
            maxdev = maxsrd

        plt.plot(diff[:, icol], alt, linetype, label='Stray Light Correction')
        plt.plot(srd, h, 'o-.', label='System Requirement')
        plt.xlim([-0.001, maxdev])
        plt.ylim([0.0, 65.0])
        plt.legend()
        plt.xlabel('Fractional Difference')
        plt.ylabel('Altitude')
        plt.title('Fractional Difference for {} nm, scene {}'.format(self._wavelength, stimulus))
        plt.savefig(os.path.join(basedir, 'stray_light_fracdiff_{}_{:1d}_{:1d}.png'.format(figure_folder, int(self._wavelength), stimulus)))

    # ------------------------------------------------------------------------------
    #           analyse_one_configuration
    # ------------------------------------------------------------------------------
    def analyse_one_configuration(self, stimulus_example: int, make_noisy_profile: bool = False):
        """
        Script to analyse the stray light preformance for a set of pre-computed BISA profiles.

        :param stimulus_example:
        :param stimulus_wavelen_index:
        :param srdvals:
        :param channel:
        :param wavelength:
        :param datestr:
        :return:
        """
        truth, angles, altitude, wavelength = self._analysis_simulator.make_l1c_stimulus_signals(stimulus_example, self._wavelength_index)     # calculate the BIA signal, convolving over the entire NOOF
        true_signal = truth['true']                                                                                                             # Get the true "desired" signal
        measured_signal, measured_error = self.calculate_noisy_signal(truth['measured'], make_noisy_profile)
        true_estimate = self._algorithm.estimate_true_profile(measured_signal, measured_error, altitude, angles)

        i = stimulus_example
        self._l1c_measured_signal[i, :, :] = measured_signal
        self._l1c_measured_error[i, :, :] = measured_error
        self._l1c_true_signal[i, :, :] = true_signal
        self._l1c_true_estimate[i, :, :] = true_estimate
        self._l1c_true_infield_stray[i, :, :] = truth['infield_stray']
        self._l1c_estimate_infield_stray[i, :, :] = self._algorithm.infield_stray_estimate
        self._l1c_true_horizontal_stray[i, :, :] = truth['horiz_stray']
        self._l1c_estimate_horizontal_stray[i, :, :] = self._algorithm.horizontal_stray_estimate
        self._l1c_true_vertical_stray[i, :, :] = truth['vertical_stray']
        self._l1c_estimate_vertical_stray[i, :, :] = self._algorithm.vertical_stray_estimate
        self._l1c_altitude[i, :] = altitude
        self._l1c_angle[i, :] = angles
        self._true_signal[i, :] = self._simulation.radiance[:, self._wavelength_index]
        self._true_altitude[i, :] = self._simulation.altitude
        self._true_angle[i, :] = self._simulation.angle
        self.make_figures(make_noisy_profile, stimulus_example)

    # ------------------------------------------------------------------------------
    #           write_results_to_netcdf
    # ------------------------------------------------------------------------------
    def write_results_to_netcdf(self, make_noisy_profile):
        """
        Writes the results to a netcdf file.
        """
        figure_folder = 'shotnoise/results' if make_noisy_profile else 'ideal/results'
        basedir = os.path.join(self._results_dir, figure_folder)
        os.makedirs(basedir, exist_ok=True)

        filename = os.path.normpath(os.path.join(basedir, 'stray_light_data_bisa_scenes_{:3d}_{:s}_{:s}.nc'.format(int(self._wavelength), self._generation_config.datestr, self._retrieval_config.datestr)))
        datavars = {}
        datavars['l1c_measured_signal'] = xr.DataArray(data=self._l1c_measured_signal, name='l1c_measured_signal', dims=["scene", "Ky", "Kx"])
        datavars['l1c_measured_error'] = xr.DataArray(data=self._l1c_measured_error, name='l1c_measured_error', dims=["scene", "Ky", "Kx"])
        datavars['l1c_true_signal'] = xr.DataArray(data=self._l1c_true_signal, name='l1c_true_signal', dims=["scene", "Ky", "Kx"])
        datavars['l1c_true_estimate'] = xr.DataArray(data=self._l1c_true_estimate, name='l1c_true_estimate', dims=["scene", "Ky", "Kx"])
        datavars['l1c_true_infield_stray'] = xr.DataArray(data=self._l1c_true_infield_stray, name='l1c_true_infield_stray', dims=["scene", "Ky", "Kx"])
        datavars['l1c_estimate_infield_stray'] = xr.DataArray(data=self._l1c_estimate_infield_stray, name='l1c_estimate_infield_stray', dims=["scene", "Ky", "Kx"])
        datavars['l1c_true_horizontal_stray'] = xr.DataArray(data=self._l1c_true_horizontal_stray, name='l1c_true_horizontal_stray', dims=["scene", "Ky", "Kx"])
        datavars['l1c_estimate_horizontal_stray'] = xr.DataArray(data=self._l1c_estimate_horizontal_stray, name='l1c_estimate_horizontal_stray', dims=["scene", "Ky", "Kx"])
        datavars['l1c_true_vertical_stray'] = xr.DataArray(data=self._l1c_true_vertical_stray, name='l1c_true_vertical_stray', dims=["scene", "Ky", "Kx"])
        datavars['l1c_estimate_vertical_stray'] = xr.DataArray(data=self._l1c_estimate_vertical_stray, name='l1c_estimate_vertical_stray', dims=["scene", "Ky", "Kx"])
        datavars['l1c_altitude'] = xr.DataArray(data=self._l1c_altitude, name='l1c_altitude', dims=["scene", "Ky"])
        datavars['l1c_angle'] = xr.DataArray(data=self._l1c_angle, name='l1c_angle', dims=["scene", "Ky"])
        datavars['full_true_signal'] = xr.DataArray(data=self._true_signal, name='full_true_signal', dims=["scene", "Ty"])
        datavars['full_altitude'] = xr.DataArray(data=self._true_altitude, name='full_altitude', dims=["scene", "Ty"])
        datavars['full_angle'] = xr.DataArray(data=self._true_angle, name='full_angle', dims=["scene", "Ty"])

        ds = xr.Dataset(data_vars=datavars)
        ds.to_netcdf(filename)
        print('Netcdf results have been written to [{}]'.format(filename))


# ------------------------------------------------------------------------------
#           main
# ------------------------------------------------------------------------------
def main():
    appname = "stray_light_removal_study_bisa_profiles"
    parser = argparse.ArgumentParser(description="BISA stray light study")
    parser.add_argument('--retrieval_date', action='store', required=True, help='The PSF entry used by the stray light removal algorithm, Typically in the format yyyy-mm-dd. Must match a wavelength entry in the psf.yaml')
    parser.add_argument('--generation_date', action='store', required=True, help='The PSF entry used by the stray light removal algorithm, Typically in the format yyyy-mm-dd. Must match a wavelength entry in the psf.yaml')
    parser.add_argument('--study', action='store', required=True, help='The name of this study. Results and figures will be written to this sub-folder ')
    parser.add_argument('--outputfolder', action='store', help='The base location of the output folder', default="./")      # C:/Users/nickl/OneDrive - University of Saskatchewan/Altius_L1PP/Stray_light_analysis")
    commandline = parser.parse_args()
    study_name = commandline.study
    retrieval_datestr = commandline.retrieval_date
    generation_datestr = commandline.generation_date
    resultsfolder = commandline.outputfolder

    resultsdir = os.path.normpath(os.path.join(resultsfolder, study_name))
    cachedir = appdirs.user_cache_dir(appname)
    print('Study:[{}]'.format(study_name))
    print('Stray light will be added to L1C profiles with PSF from [{}]'.format(generation_datestr))
    print('Stray light will be removed from L1C profiles with PSF from [{}]'.format(retrieval_datestr))
    print('Results will be written to folder [{}]'.format(resultsdir))

    channel_wavelen = ['uv', 'uv', 'uv', 'vis', 'vis', 'vis']
    wavelen = [300.0, 300.0, 350.0, 550.0, 550.0, 550.0]
    simulation = SimulationRadianceDataBISA(case=0)
    make_noisy_profile = True

    # stimulus_wavelen_index = 5
    # stimulus_example = 0
    # channel = channel_wavelen[stimulus_wavelen_index]
    # pwave = wavelen[stimulus_wavelen_index]
    # config = AltiusConfigurationKey(channel=channel, wavelength=pwave, datestr='2020-10-20')
    # study = StrayLightRemovalStudyBISAProfiles( stimulus_wavelen_index, config, simulation )
    # study.analyse_one_configuration(stimulus_example, make_noisy_profile=make_noisy_profile)

    for stimulus_wavelen_index in range(0, 6):
        channel = channel_wavelen[stimulus_wavelen_index]
        pwave = wavelen[stimulus_wavelen_index]
        retrieval_config = AltiusConfigurationKey(channel=channel, wavelength=pwave, datestr=retrieval_datestr)
        generation_config = AltiusConfigurationKey(channel=channel, wavelength=pwave, datestr=generation_datestr)
        study = StrayLightRemovalStudyBISAProfiles(stimulus_wavelen_index, retrieval_config, generation_config, simulation, cache_dir=cachedir, results_dir=resultsdir)

        for stimulus_example in range(16):
            print('Processing stimulus example {}, wavelength index {}'.format(stimulus_example, stimulus_wavelen_index))
            study.analyse_one_configuration(stimulus_example, make_noisy_profile=make_noisy_profile)

        study.write_results_to_netcdf(make_noisy_profile)
    print("Done")


if __name__ == '__main__':
    main()
